/*
 * SetupOps.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using Eagle._Attributes;
using Eagle._Components.Public;
using Eagle._Containers.Public;
using Microsoft.Win32;

namespace Eagle._Components.Private
{
    [ObjectId("bd1dfe4f-116f-4780-9bbf-4c8c80d94873")]
    internal static class SetupOps
    {
        #region Private Constants
#if NATIVE && WINDOWS
        private static readonly string mutexName =
            GlobalState.GetPackageName(null, "_Setup", false);

        private static readonly string globalMutexName =
            "Global\\" + mutexName;
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static readonly string keyName =
            GlobalState.GetPackageName("Software\\", null, false);

        private const string assemblyValueName = "Assembly";

        private const string appIdValueName = "AppId";

        private const string compileInfoValueName = "CompileInfo";

        private const string pathValueName = "Path";
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Data
#if NATIVE && WINDOWS
        //
        // NOTE: Used to serialize access to the mutex handles below.
        //
        private static readonly object syncRoot = new object();

        //
        // NOTE: *NEVER CLOSED* Prevents setup from installing.
        //
        private static IntPtr globalMutex = IntPtr.Zero;
        private static IntPtr mutex = IntPtr.Zero;
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static string GetPath(
            Version version
            )
        {
            return GetPath(Registry.LocalMachine, version);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static string GetPath(
            RegistryKey rootKey,
            Version version
            )
        {
            RegistryKey key = null;

            try
            {
                if ((rootKey != null) &&
                    !String.IsNullOrEmpty(keyName) &&
                    (version != null))
                {
                    key = rootKey.OpenSubKey(String.Format("{0}\\{1}",
                        keyName, version));

                    if (key != null)
                        return key.GetValue(pathValueName) as string;
                }
            }
            catch
            {
                // do nothing.
            }
            finally
            {
                if (key != null)
                {
                    key.Close();
                    key = null;
                }
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode GetInstances(
            ref Result result
            )
        {
            return GetInstances(Registry.LocalMachine, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode GetInstances(
            RegistryKey rootKey,
            ref Result result
            )
        {
            RegistryKey key = null;

            try
            {
                if ((rootKey != null) &&
                    !String.IsNullOrEmpty(keyName))
                {
                    key = rootKey.OpenSubKey(keyName);

                    if (key != null)
                    {
                        StringList list = new StringList();

                        foreach (string subKeyName in key.GetSubKeyNames())
                        {
                            StringList subList =
                                new StringList("Name", subKeyName);

                            RegistryKey subKey = null;

                            try
                            {
                                subKey = key.OpenSubKey(subKeyName);

                                if (subKey != null)
                                {
                                    subList.Add(appIdValueName,
                                        subKey.GetValue(appIdValueName) as string);
                                    subList.Add(pathValueName,
                                        subKey.GetValue(pathValueName) as string);
                                    subList.Add(assemblyValueName,
                                        subKey.GetValue(assemblyValueName) as string);
                                    subList.Add(compileInfoValueName,
                                        subKey.GetValue(compileInfoValueName) as string);
                                }
                                else
                                {
                                    subList.Add("Error", String.Format(
                                        "could not open sub-key \"{0}\" of registry key \"{1}\"",
                                        subKeyName, key.Name));
                                }
                            }
                            catch (Exception e)
                            {
                                subList.Add("Exception", e.ToString());
                                subList.Add("Error", String.Format(
                                    "could not open sub-key \"{0}\" of registry key \"{1}\"",
                                    subKeyName, key.Name));
                            }
                            finally
                            {
                                if (subKey != null)
                                {
                                    subKey.Close();
                                    subKey = null;
                                }
                            }

                            list.Add(subList.ToString());
                        }

                        result = list;
                        return ReturnCode.Ok;
                    }
                    else
                    {
                        result = "no setup instances found";
                    }
                }
                else
                {
                    result = "invalid registry key name";
                }
            }
            catch (Exception e)
            {
                result = e;
            }
            finally
            {
                if (key != null)
                {
                    key.Close();
                    key = null;
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if NATIVE && WINDOWS
        public static void CreateMutexes()
        {
            ReturnCode mutexCode;
            Result mutexError = null;

            mutexCode = CreateMutexes(ref mutexError);

            if (mutexCode != ReturnCode.Ok)
                DebugOps.Complain(null, mutexCode, mutexError);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode CreateMutexes(
            ref Result error
            )
        {
            if (PlatformOps.IsWindowsOperatingSystem())
            {
                if (!String.IsNullOrEmpty(globalMutexName))
                {
                    if (!String.IsNullOrEmpty(mutexName))
                    {
                        try
                        {
                            lock (syncRoot)
                            {
                                //
                                // NOTE: *NEVER CLOSED* This should prevent our setup
                                //       from running while the library is in use
                                //       (even if the setup is running in a different
                                //       user session).
                                //
                                if (globalMutex == IntPtr.Zero)
                                    globalMutex = NativeOps.UnsafeNativeMethods.CreateMutex(
                                        IntPtr.Zero, true, globalMutexName);

                                if (globalMutex == IntPtr.Zero)
                                    DebugOps.Complain(ReturnCode.Error, String.Format(
                                        "could not create global mutex \"{0}\"",
                                        globalMutexName));

                                //
                                // NOTE: *NEVER CLOSED* This should prevent our setup
                                //       from running while the library is in use.
                                //
                                if (mutex == IntPtr.Zero)
                                    mutex = NativeOps.UnsafeNativeMethods.CreateMutex(
                                        IntPtr.Zero, true, mutexName);

                                if (mutex == IntPtr.Zero)
                                    DebugOps.Complain(ReturnCode.Error, String.Format(
                                        "could not create mutex \"{0}\"", mutexName));
                            }

                            return ReturnCode.Ok;
                        }
                        catch (Exception e)
                        {
                            error = e;
                        }
                    }
                    else
                    {
                        error = "invalid mutex name";
                    }
                }
                else
                {
                    error = "invalid global mutex name";
                }
            }
            else
            {
                error = "not supported on this operating system";
            }

            return ReturnCode.Error;
        }
#endif
    }
}
