/*
 * Default.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Text;
using System.Threading;
using Eagle._Attributes;
using Eagle._Components.Private;
using Eagle._Components.Public;
using Eagle._Components.Public.Delegates;
using Eagle._Constants;
using Eagle._Containers.Private;
using Eagle._Containers.Public;
using Eagle._Interfaces.Private;
using Eagle._Interfaces.Public;

#if !CONSOLE
using ConsoleColor = Eagle._Components.Public.ConsoleColor;
#endif

//
// TODO: Load all the strings used in this class from resources.
//
namespace Eagle._Hosts
{
    [ObjectId("4a02c60b-5b7b-4eb8-873e-eb6860ba3973")]
    public abstract class Default :
#if ISOLATED_INTERPRETERS || ISOLATED_PLUGINS
        ScriptMarshalByRefObject,
#endif
        IHost, IDisposable
    {
        #region Protected Constants
        protected static readonly BindingFlags HostPropertyBindingFlags =
            BindingFlags.IgnoreCase | BindingFlags.Instance |
            BindingFlags.Public | BindingFlags.NonPublic |
            BindingFlags.SetProperty;

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // NOTE: These are the "default" beep values, per MSDN.
        //
        protected internal static readonly int BeepFrequency = 800;
        protected internal static readonly int BeepDuration = 200;

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Section Names
        protected static readonly string HeaderSectionName = "Header";
        protected static readonly string FooterSectionName = "Footer";
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Box Names
        protected static readonly string ArgumentInfoBoxName = "ArgumentInfo";
        protected static readonly string CallFrameInfoBoxName = "CallFrameInfo";
        protected static readonly string CallStackInfoBoxName = "CallStackInfo";
        protected static readonly string DebuggerInfoBoxName = "DebuggerInfo";
        protected static readonly string FlagInfoBoxName = "FlagInfo";
        protected static readonly string HostInfoBoxName = "HostInfo";
        protected static readonly string InterpreterInfoBoxName = "InterpreterInfo";
        protected static readonly string EngineInfoBoxName = "EngineInfo";
        protected static readonly string EntityInfoBoxName = "EntityInfo";
        protected static readonly string StackInfoBoxName = "StackInfo";
        protected static readonly string ControlInfoBoxName = "ControlInfo";
        protected static readonly string TestInfoBoxName = "TestInfo";
        protected static readonly string TraceInfoBoxName = "TraceInfo";
        protected static readonly string VariableInfoBoxName = "VariableInfo";
        protected static readonly string ObjectInfoBoxName = "ObjectInfo";
        protected static readonly string ComplaintInfoBoxName = "ComplaintInfo";
        protected static readonly string HistoryInfoBoxName = "HistoryInfo";
        protected static readonly string CustomInfoBoxName = "CustomInfo";
        protected static readonly string ResultInfoBoxName = "ResultInfo";
        protected static readonly string PreviousResultInfoBoxName = "PreviousResultInfo";
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Public Constants
        public static readonly string BoxColorPrefix = "Box";

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static readonly char NoSeparator = char.MinValue;

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static readonly bool DefaultCanExit = true;
        public static readonly bool DefaultCanForceExit = true;
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Constructors
        private Default()
        {
            kind = IdentifierKind.Host;
            id = AttributeOps.GetObjectId(this); /* VIRTUAL: Id of the deepest derived class. */

            //
            // NOTE: Set the defaults for the exit-related properties.
            //
            canExit = DefaultCanExit;
            canForceExit = DefaultCanForceExit;

            //
            // NOTE: Set the default display constraints for each section of
            //       the header.
            //
            if (sectionSizes == null)
                sectionSizes = new Dictionary<HeaderFlags, HostFlags>();

            if (sectionSizes.Count == 0)
            {
                sectionSizes.Add(HeaderFlags.StopPrompt, HostFlags.CompactSize);
                sectionSizes.Add(HeaderFlags.GoPrompt, HostFlags.CompactSize);
                sectionSizes.Add(HeaderFlags.AnnouncementInfo, HostFlags.FullSize);
#if DEBUGGER
                sectionSizes.Add(HeaderFlags.DebuggerInfo, HostFlags.FullSize);
#endif
                sectionSizes.Add(HeaderFlags.EngineInfo, HostFlags.FullSize);
                sectionSizes.Add(HeaderFlags.ControlInfo, HostFlags.FullSize);
                sectionSizes.Add(HeaderFlags.EntityInfo, HostFlags.JumboSize);
                sectionSizes.Add(HeaderFlags.StackInfo, HostFlags.JumboSize);
                sectionSizes.Add(HeaderFlags.FlagInfo, HostFlags.FullSize);
                sectionSizes.Add(HeaderFlags.HostInfo, HostFlags.SuperJumboSize);
                sectionSizes.Add(HeaderFlags.ArgumentInfo, HostFlags.ZeroSize);
                sectionSizes.Add(HeaderFlags.TestInfo, HostFlags.JumboSize);
                sectionSizes.Add(HeaderFlags.TokenInfo, HostFlags.ZeroSize);
                sectionSizes.Add(HeaderFlags.TraceInfo, HostFlags.ZeroSize);
                sectionSizes.Add(HeaderFlags.VariableInfo, HostFlags.JumboSize);
                sectionSizes.Add(HeaderFlags.ObjectInfo, HostFlags.JumboSize);
                sectionSizes.Add(HeaderFlags.InterpreterInfo, HostFlags.JumboSize);
                sectionSizes.Add(HeaderFlags.ComplaintInfo, HostFlags.ZeroSize);
                sectionSizes.Add(HeaderFlags.CustomInfo, HostFlags.ZeroSize);
                sectionSizes.Add(HeaderFlags.ResultInfo, HostFlags.ZeroSize);
#if HISTORY
                sectionSizes.Add(HeaderFlags.HistoryInfo, HostFlags.JumboSize);
#endif
                sectionSizes.Add(HeaderFlags.CallStack, HostFlags.CompactSize);
                sectionSizes.Add(HeaderFlags.CallStackInfo, HostFlags.FullSize);
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Protected Constructors
        protected Default(
            IHostData hostData
            )
            : this()
        {
            if (hostData != null)
            {
                name = hostData.Name;
                group = hostData.Group;
                description = hostData.Description;
                clientData = hostData.ClientData;

                //
                // NOTE: Use the profile provided by the caller, if any.
                //
                profile = hostData.Profile;

                //
                // NOTE: Allow colors to be disabled.
                //
                noColor = hostData.NoColor;

                //
                // NOTE: Allow setting the host title to be disabled.
                //
                noTitle = hostData.NoTitle;

                //
                // NOTE: Allow setting the host icon to be disabled.
                //
                noIcon = hostData.NoIcon;

                //
                // NOTE: Allow loading of the host profile to be disabled.
                //
                noProfile = hostData.NoProfile;

                //
                // NOTE: Allow the script cancellation interface to be disabled.
                //
                noCancel = hostData.NoCancel;
            }

            //
            // NOTE: Initialize the system default colors for the various elements
            //       of the user-interface.
            //
            InitializeColors();
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Protected Properties
        #region Protected Section Properties
        private Dictionary<HeaderFlags, HostFlags> sectionSizes;
        protected virtual Dictionary<HeaderFlags, HostFlags> SectionSizes
        {
            get { return sectionSizes; }
            set { sectionSizes = value; }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Protected Call Frame Properties
        private string variableCallFrameSuffix = null;
        protected virtual string VariableCallFrameSuffix
        {
            get { return variableCallFrameSuffix; }
            set { variableCallFrameSuffix = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private char variableCallFrameSeparator = '*';
        protected virtual char VariableCallFrameSeparator
        {
            get { return variableCallFrameSeparator; }
            set { variableCallFrameSeparator = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string nullCallFrameTypeName = _String.Null;
        protected virtual string NullCallFrameTypeName
        {
            get { return nullCallFrameTypeName; }
            set { nullCallFrameTypeName = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string variableCallFrameTypeName = "vars";
        protected virtual string VariableCallFrameTypeName
        {
            get { return variableCallFrameTypeName; }
            set { variableCallFrameTypeName = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string currentCallFrameTypeName = "curr";
        protected virtual string CurrentCallFrameTypeName
        {
            get { return currentCallFrameTypeName; }
            set { currentCallFrameTypeName = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string procedureCallFrameTypeName = "proc";
        protected virtual string ProcedureCallFrameTypeName
        {
            get { return procedureCallFrameTypeName; }
            set { procedureCallFrameTypeName = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string globalCallFrameTypeName = "glob";
        protected virtual string GlobalCallFrameTypeName
        {
            get { return globalCallFrameTypeName; }
            set { globalCallFrameTypeName = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string lambdaCallFrameTypeName = "lamb";
        protected virtual string LambdaCallFrameTypeName
        {
            get { return lambdaCallFrameTypeName; }
            set { lambdaCallFrameTypeName = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string scopeCallFrameTypeName = "scop";
        protected virtual string ScopeCallFrameTypeName
        {
            get { return scopeCallFrameTypeName; }
            set { scopeCallFrameTypeName = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string aliasCallFrameTypeName = "alis";
        protected virtual string AliasCallFrameTypeName
        {
            get { return aliasCallFrameTypeName; }
            set { aliasCallFrameTypeName = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string uplevelCallFrameTypeName = "uplv";
        protected virtual string UplevelCallFrameTypeName
        {
            get { return uplevelCallFrameTypeName; }
            set { uplevelCallFrameTypeName = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string trackingCallFrameTypeName = "trck";
        protected virtual string TrackingCallFrameTypeName
        {
            get { return trackingCallFrameTypeName; }
            set { trackingCallFrameTypeName = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string unknownCallFrameTypeName = "unkn";
        protected virtual string UnknownCallFrameTypeName
        {
            get { return unknownCallFrameTypeName; }
            set { unknownCallFrameTypeName = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string linkedCallFrameTypeName = "link";
        protected virtual string LinkedCallFrameTypeName
        {
            get { return linkedCallFrameTypeName; }
            set { linkedCallFrameTypeName = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // NOTE: Yes, this normally returns only whitespace.
        //
        private string normalCallFrameTypeName = "    ";
        protected virtual string NormalCallFrameTypeName
        {
            get { return normalCallFrameTypeName; }
            set { normalCallFrameTypeName = value; }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Protected Formatting Properties
        private bool debug = false;
        protected virtual bool Debug
        {
            get { return debug; }
            set { debug = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool replaceNewLines = true;
        protected internal virtual bool ReplaceNewLines
        {
            get { return replaceNewLines; }
            set { replaceNewLines = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool ellipsis = true;
        protected internal virtual bool Ellipsis
        {
            get { return ellipsis; }
            set { ellipsis = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool exceptions = false;
        protected internal virtual bool Exceptions
        {
            get { return exceptions; }
            set { exceptions = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private int historyLimit = 20;
        protected virtual int HistoryLimit
        {
            get { return historyLimit; }
            set { historyLimit = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private int callStackLimit = 20;
        protected virtual int CallStackLimit
        {
            get { return callStackLimit; }
            set { callStackLimit = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private int sectionsPerRow = 0;
        protected virtual int SectionsPerRow
        {
            get { return sectionsPerRow; }
            set { sectionsPerRow = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string nameValueFormat = "{0}:" + Characters.Space + "{1}";
        protected virtual string NameValueFormat
        {
            get { return nameValueFormat; }
            set { nameValueFormat = value; }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Protected Prompt Properties
        private string stopPrompt = "[stop]";
        protected virtual string StopPrompt
        {
            get { return stopPrompt; }
            set { stopPrompt = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string goPrompt = "[go]";
        protected virtual string GoPrompt
        {
            get { return goPrompt; }
            set { goPrompt = value; }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Protected Output Properties
        private OutputStyle outputStyle = OutputStyle.Default;
        protected internal virtual OutputStyle OutputStyle // TODO: Make this part of IHost?
        {
            get { return outputStyle; }
            set { outputStyle = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Protected Window Properties
        private int hostLeft = 0;
        protected virtual int HostLeft
        {
            get { return hostLeft; }
            set { hostLeft = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private int hostTop = 0;
        protected virtual int HostTop
        {
            get { return hostTop; }
            set { hostTop = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private int windowWidth = 80;
        protected virtual int WindowWidth
        {
            get { return windowWidth; }
            set { windowWidth = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private int windowHeight = 25;
        protected virtual int WindowHeight
        {
            get { return windowHeight; }
            set { windowHeight = value; }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Protected Content Area Properties
        private int contentMargin = 0;
        protected virtual int ContentMargin
        {
            get { return contentMargin; }
            set { contentMargin = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private int contentWidth = Width.Invalid;
        protected virtual int ContentWidth
        {
            get { return (contentWidth != Width.Invalid) ? contentWidth : WindowWidth - 3; }
            set { contentWidth = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private int contentThreshold = 20;
        protected virtual int ContentThreshold
        {
            get { return contentThreshold; }
            set { contentThreshold = value; }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Protected Box Properties
#if WINDOWS
        private int boxCharacterSet = Characters.BoxCharacterSets.Count - 1;
#else
        private int boxCharacterSet = 0;
#endif
        protected internal virtual int BoxCharacterSet
        {
            get { return boxCharacterSet; }
            set { boxCharacterSet = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private StringList boxCharacterSets = Characters.BoxCharacterSets;
        protected internal virtual StringList BoxCharacterSets
        {
            get { return boxCharacterSets; }
            set { boxCharacterSets = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private int boxWidth = Width.Invalid;
        protected virtual int BoxWidth
        {
            get { return boxWidth; }
            set { boxWidth = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private int boxMargin = 2;
        protected virtual int BoxMargin
        {
            get { return boxMargin; }
            set { boxMargin = value; }
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Protected Color Properties
        #region General & Section Colors
        #region General Colors
        private ConsoleColor bannerForegroundColor;
        protected virtual ConsoleColor BannerForegroundColor
        {
            get { return bannerForegroundColor; }
            set { bannerForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor bannerBackgroundColor;
        protected virtual ConsoleColor BannerBackgroundColor
        {
            get { return bannerBackgroundColor; }
            set { bannerBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor boxForegroundColor;
        protected virtual ConsoleColor BoxForegroundColor
        {
            get { return boxForegroundColor; }
            set { boxForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor boxBackgroundColor;
        protected virtual ConsoleColor BoxBackgroundColor
        {
            get { return boxBackgroundColor; }
            set { boxBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor debugForegroundColor;
        protected virtual ConsoleColor DebugForegroundColor
        {
            get { return debugForegroundColor; }
            set { debugForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor debugBackgroundColor;
        protected virtual ConsoleColor DebugBackgroundColor
        {
            get { return debugBackgroundColor; }
            set { debugBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor errorForegroundColor;
        protected virtual ConsoleColor ErrorForegroundColor
        {
            get { return errorForegroundColor; }
            set { errorForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor errorBackgroundColor;
        protected virtual ConsoleColor ErrorBackgroundColor
        {
            get { return errorBackgroundColor; }
            set { errorBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor defaultForegroundColor;
        protected virtual ConsoleColor DefaultForegroundColor
        {
            get { return defaultForegroundColor; }
            set { defaultForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor defaultBackgroundColor;
        protected virtual ConsoleColor DefaultBackgroundColor
        {
            get { return defaultBackgroundColor; }
            set { defaultBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor footerForegroundColor;
        protected virtual ConsoleColor FooterForegroundColor
        {
            get { return footerForegroundColor; }
            set { footerForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor footerBackgroundColor;
        protected virtual ConsoleColor FooterBackgroundColor
        {
            get { return footerBackgroundColor; }
            set { footerBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor headerForegroundColor;
        protected virtual ConsoleColor HeaderForegroundColor
        {
            get { return headerForegroundColor; }
            set { headerForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor headerBackgroundColor;
        protected virtual ConsoleColor HeaderBackgroundColor
        {
            get { return headerBackgroundColor; }
            set { headerBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor helpForegroundColor;
        protected virtual ConsoleColor HelpForegroundColor
        {
            get { return helpForegroundColor; }
            set { helpForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor helpBackgroundColor;
        protected virtual ConsoleColor HelpBackgroundColor
        {
            get { return helpBackgroundColor; }
            set { helpBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor helpItemForegroundColor;
        protected virtual ConsoleColor HelpItemForegroundColor
        {
            get { return helpItemForegroundColor; }
            set { helpItemForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor helpItemBackgroundColor;
        protected virtual ConsoleColor HelpItemBackgroundColor
        {
            get { return helpItemBackgroundColor; }
            set { helpItemBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor legalForegroundColor;
        protected virtual ConsoleColor LegalForegroundColor
        {
            get { return legalForegroundColor; }
            set { legalForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor legalBackgroundColor;
        protected virtual ConsoleColor LegalBackgroundColor
        {
            get { return legalBackgroundColor; }
            set { legalBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor resultForegroundColor;
        protected virtual ConsoleColor ResultForegroundColor
        {
            get { return resultForegroundColor; }
            set { resultForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor resultBackgroundColor;
        protected virtual ConsoleColor ResultBackgroundColor
        {
            get { return resultBackgroundColor; }
            set { resultBackgroundColor = value; }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Section Colors
        private ConsoleColor announcementInfoForegroundColor;
        protected virtual ConsoleColor AnnouncementInfoForegroundColor
        {
            get { return announcementInfoForegroundColor; }
            set { announcementInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor announcementInfoBackgroundColor;
        protected virtual ConsoleColor AnnouncementInfoBackgroundColor
        {
            get { return announcementInfoBackgroundColor; }
            set { announcementInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor argumentInfoForegroundColor;
        protected virtual ConsoleColor ArgumentInfoForegroundColor
        {
            get { return argumentInfoForegroundColor; }
            set { argumentInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor argumentInfoBackgroundColor;
        protected virtual ConsoleColor ArgumentInfoBackgroundColor
        {
            get { return argumentInfoBackgroundColor; }
            set { argumentInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor callFrameInfoForegroundColor;
        protected virtual ConsoleColor CallFrameInfoForegroundColor
        {
            get { return callFrameInfoForegroundColor; }
            set { callFrameInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor callFrameInfoBackgroundColor;
        protected virtual ConsoleColor CallFrameInfoBackgroundColor
        {
            get { return callFrameInfoBackgroundColor; }
            set { callFrameInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor callStackInfoForegroundColor;
        protected virtual ConsoleColor CallStackInfoForegroundColor
        {
            get { return callStackInfoForegroundColor; }
            set { callStackInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor callStackInfoBackgroundColor;
        protected virtual ConsoleColor CallStackInfoBackgroundColor
        {
            get { return callStackInfoBackgroundColor; }
            set { callStackInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor complaintInfoForegroundColor;
        protected virtual ConsoleColor ComplaintInfoForegroundColor
        {
            get { return complaintInfoForegroundColor; }
            set { complaintInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor complaintInfoBackgroundColor;
        protected virtual ConsoleColor ComplaintInfoBackgroundColor
        {
            get { return complaintInfoBackgroundColor; }
            set { complaintInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor controlInfoForegroundColor;
        protected virtual ConsoleColor ControlInfoForegroundColor
        {
            get { return controlInfoForegroundColor; }
            set { controlInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor controlInfoBackgroundColor;
        protected virtual ConsoleColor ControlInfoBackgroundColor
        {
            get { return controlInfoBackgroundColor; }
            set { controlInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor customInfoForegroundColor;
        protected virtual ConsoleColor CustomInfoForegroundColor
        {
            get { return customInfoForegroundColor; }
            set { customInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor customInfoBackgroundColor;
        protected virtual ConsoleColor CustomInfoBackgroundColor
        {
            get { return customInfoBackgroundColor; }
            set { customInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor debuggerInfoForegroundColor;
        protected virtual ConsoleColor DebuggerInfoForegroundColor
        {
            get { return debuggerInfoForegroundColor; }
            set { debuggerInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor debuggerInfoBackgroundColor;
        protected virtual ConsoleColor DebuggerInfoBackgroundColor
        {
            get { return debuggerInfoBackgroundColor; }
            set { debuggerInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor engineInfoForegroundColor;
        protected virtual ConsoleColor EngineInfoForegroundColor
        {
            get { return engineInfoForegroundColor; }
            set { engineInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor engineInfoBackgroundColor;
        protected virtual ConsoleColor EngineInfoBackgroundColor
        {
            get { return engineInfoBackgroundColor; }
            set { engineInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor entityInfoForegroundColor;
        protected virtual ConsoleColor EntityInfoForegroundColor
        {
            get { return entityInfoForegroundColor; }
            set { entityInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor entityInfoBackgroundColor;
        protected virtual ConsoleColor EntityInfoBackgroundColor
        {
            get { return entityInfoBackgroundColor; }
            set { entityInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor flagInfoForegroundColor;
        protected virtual ConsoleColor FlagInfoForegroundColor
        {
            get { return flagInfoForegroundColor; }
            set { flagInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor flagInfoBackgroundColor;
        protected virtual ConsoleColor FlagInfoBackgroundColor
        {
            get { return flagInfoBackgroundColor; }
            set { flagInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor historyInfoForegroundColor;
        protected virtual ConsoleColor HistoryInfoForegroundColor
        {
            get { return historyInfoForegroundColor; }
            set { historyInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor historyInfoBackgroundColor;
        protected virtual ConsoleColor HistoryInfoBackgroundColor
        {
            get { return historyInfoBackgroundColor; }
            set { historyInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor hostInfoForegroundColor;
        protected virtual ConsoleColor HostInfoForegroundColor
        {
            get { return hostInfoForegroundColor; }
            set { hostInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor hostInfoBackgroundColor;
        protected virtual ConsoleColor HostInfoBackgroundColor
        {
            get { return hostInfoBackgroundColor; }
            set { hostInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor interpreterInfoForegroundColor;
        protected virtual ConsoleColor InterpreterInfoForegroundColor
        {
            get { return interpreterInfoForegroundColor; }
            set { interpreterInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor interpreterInfoBackgroundColor;
        protected virtual ConsoleColor InterpreterInfoBackgroundColor
        {
            get { return interpreterInfoBackgroundColor; }
            set { interpreterInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor objectInfoForegroundColor;
        protected virtual ConsoleColor ObjectInfoForegroundColor
        {
            get { return objectInfoForegroundColor; }
            set { objectInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor objectInfoBackgroundColor;
        protected virtual ConsoleColor ObjectInfoBackgroundColor
        {
            get { return objectInfoBackgroundColor; }
            set { objectInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor testInfoForegroundColor;
        protected virtual ConsoleColor TestInfoForegroundColor
        {
            get { return testInfoForegroundColor; }
            set { testInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor testInfoBackgroundColor;
        protected virtual ConsoleColor TestInfoBackgroundColor
        {
            get { return testInfoBackgroundColor; }
            set { testInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor tokenInfoForegroundColor;
        protected virtual ConsoleColor TokenInfoForegroundColor
        {
            get { return tokenInfoForegroundColor; }
            set { tokenInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor tokenInfoBackgroundColor;
        protected virtual ConsoleColor TokenInfoBackgroundColor
        {
            get { return tokenInfoBackgroundColor; }
            set { tokenInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor traceInfoForegroundColor;
        protected virtual ConsoleColor TraceInfoForegroundColor
        {
            get { return traceInfoForegroundColor; }
            set { traceInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor traceInfoBackgroundColor;
        protected virtual ConsoleColor TraceInfoBackgroundColor
        {
            get { return traceInfoBackgroundColor; }
            set { traceInfoBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor variableInfoForegroundColor;
        protected virtual ConsoleColor VariableInfoForegroundColor
        {
            get { return variableInfoForegroundColor; }
            set { variableInfoForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor variableInfoBackgroundColor;
        protected virtual ConsoleColor VariableInfoBackgroundColor
        {
            get { return variableInfoBackgroundColor; }
            set { variableInfoBackgroundColor = value; }
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Return Code Colors
        private ConsoleColor okResultForegroundColor;
        protected virtual ConsoleColor OkResultForegroundColor
        {
            get { return okResultForegroundColor; }
            set { okResultForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor okResultBackgroundColor;
        protected virtual ConsoleColor OkResultBackgroundColor
        {
            get { return okResultBackgroundColor; }
            set { okResultBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor otherOkResultForegroundColor;
        protected virtual ConsoleColor OtherOkResultForegroundColor
        {
            get { return otherOkResultForegroundColor; }
            set { otherOkResultForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor otherOkResultBackgroundColor;
        protected virtual ConsoleColor OtherOkResultBackgroundColor
        {
            get { return otherOkResultBackgroundColor; }
            set { otherOkResultBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor otherErrorResultForegroundColor;
        protected virtual ConsoleColor OtherErrorResultForegroundColor
        {
            get { return otherErrorResultForegroundColor; }
            set { otherErrorResultForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor otherErrorResultBackgroundColor;
        protected virtual ConsoleColor OtherErrorResultBackgroundColor
        {
            get { return otherErrorResultBackgroundColor; }
            set { otherErrorResultBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor errorResultForegroundColor;
        protected virtual ConsoleColor ErrorResultForegroundColor
        {
            get { return errorResultForegroundColor; }
            set { errorResultForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor errorResultBackgroundColor;
        protected virtual ConsoleColor ErrorResultBackgroundColor
        {
            get { return errorResultBackgroundColor; }
            set { errorResultBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor returnResultForegroundColor;
        protected virtual ConsoleColor ReturnResultForegroundColor
        {
            get { return returnResultForegroundColor; }
            set { returnResultForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor returnResultBackgroundColor;
        protected virtual ConsoleColor ReturnResultBackgroundColor
        {
            get { return returnResultBackgroundColor; }
            set { returnResultBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor breakResultForegroundColor;
        protected virtual ConsoleColor BreakResultForegroundColor
        {
            get { return breakResultForegroundColor; }
            set { breakResultForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor breakResultBackgroundColor;
        protected virtual ConsoleColor BreakResultBackgroundColor
        {
            get { return breakResultBackgroundColor; }
            set { breakResultBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor continueResultForegroundColor;
        protected virtual ConsoleColor ContinueResultForegroundColor
        {
            get { return continueResultForegroundColor; }
            set { continueResultForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor continueResultBackgroundColor;
        protected virtual ConsoleColor ContinueResultBackgroundColor
        {
            get { return continueResultBackgroundColor; }
            set { continueResultBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor nullResultForegroundColor;
        protected virtual ConsoleColor NullResultForegroundColor
        {
            get { return nullResultForegroundColor; }
            set { nullResultForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor nullResultBackgroundColor;
        protected virtual ConsoleColor NullResultBackgroundColor
        {
            get { return nullResultBackgroundColor; }
            set { nullResultBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor emptyResultForegroundColor;
        protected virtual ConsoleColor EmptyResultForegroundColor
        {
            get { return emptyResultForegroundColor; }
            set { emptyResultForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor emptyResultBackgroundColor;
        protected virtual ConsoleColor EmptyResultBackgroundColor
        {
            get { return emptyResultBackgroundColor; }
            set { emptyResultBackgroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor unknownResultForegroundColor;
        protected virtual ConsoleColor UnknownResultForegroundColor
        {
            get { return unknownResultForegroundColor; }
            set { unknownResultForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor unknownResultBackgroundColor;
        protected virtual ConsoleColor UnknownResultBackgroundColor
        {
            get { return unknownResultBackgroundColor; }
            set { unknownResultBackgroundColor = value; }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Call Frame Colors
        private ConsoleColor nullCallFrameForegroundColor;
        protected virtual ConsoleColor NullCallFrameForegroundColor
        {
            get { return nullCallFrameForegroundColor; }
            set { nullCallFrameForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor unknownCallFrameForegroundColor;
        protected virtual ConsoleColor UnknownCallFrameForegroundColor
        {
            get { return unknownCallFrameForegroundColor; }
            set { unknownCallFrameForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor variableCallFrameForegroundColor;
        protected virtual ConsoleColor VariableCallFrameForegroundColor
        {
            get { return variableCallFrameForegroundColor; }
            set { variableCallFrameForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor uplevelCallFrameForegroundColor;
        protected virtual ConsoleColor UplevelCallFrameForegroundColor
        {
            get { return uplevelCallFrameForegroundColor; }
            set { uplevelCallFrameForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor trackingCallFrameForegroundColor;
        protected virtual ConsoleColor TrackingCallFrameForegroundColor
        {
            get { return trackingCallFrameForegroundColor; }
            set { trackingCallFrameForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor currentCallFrameForegroundColor;
        protected virtual ConsoleColor CurrentCallFrameForegroundColor
        {
            get { return currentCallFrameForegroundColor; }
            set { currentCallFrameForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor procecureCallFrameForegroundColor;
        protected virtual ConsoleColor ProcecureCallFrameForegroundColor
        {
            get { return procecureCallFrameForegroundColor; }
            set { procecureCallFrameForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor lambdaCallFrameForegroundColor;
        protected virtual ConsoleColor LambdaCallFrameForegroundColor
        {
            get { return lambdaCallFrameForegroundColor; }
            set { lambdaCallFrameForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor scopeCallFrameForegroundColor;
        protected virtual ConsoleColor ScopeCallFrameForegroundColor
        {
            get { return scopeCallFrameForegroundColor; }
            set { scopeCallFrameForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor aliasCallFrameForegroundColor;
        protected virtual ConsoleColor AliasCallFrameForegroundColor
        {
            get { return aliasCallFrameForegroundColor; }
            set { aliasCallFrameForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor globalCallFrameForegroundColor;
        protected virtual ConsoleColor GlobalCallFrameForegroundColor
        {
            get { return globalCallFrameForegroundColor; }
            set { globalCallFrameForegroundColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor linkedCallFrameForegroundColor;
        protected virtual ConsoleColor LinkedCallFrameForegroundColor
        {
            get { return linkedCallFrameForegroundColor; }
            set { linkedCallFrameForegroundColor = value; }
        }
        #endregion
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Color Helper Methods
        #region Color Initialization Methods
        protected virtual ConsoleColor GetColor(
            ConsoleColor color
            )
        {
            //
            // BUGFIX: When the NoColor property is true, disable all colors;
            //         otherwise, just return the color passed in by the caller.
            //
            return NoColor ? _ConsoleColor.None : color;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected internal virtual void InitializeColors()
        {
            #region General Colors
            bannerForegroundColor = GetColor(ConsoleColor.White);
            bannerBackgroundColor = GetColor(_ConsoleColor.None);
            boxForegroundColor = GetColor(ConsoleColor.White);
            boxBackgroundColor = GetColor(ConsoleColor.DarkGreen);
            debugForegroundColor = GetColor(ConsoleColor.White);
            debugBackgroundColor = GetColor(_ConsoleColor.None);
            errorForegroundColor = GetColor(ConsoleColor.Red);
            errorBackgroundColor = GetColor(ConsoleColor.White);
            defaultForegroundColor = GetColor(_ConsoleColor.None);
            defaultBackgroundColor = GetColor(_ConsoleColor.None);
            footerForegroundColor = GetColor(ConsoleColor.DarkYellow);
            footerBackgroundColor = GetColor(_ConsoleColor.None);
            headerForegroundColor = GetColor(ConsoleColor.DarkYellow);
            headerBackgroundColor = GetColor(_ConsoleColor.None);
            helpForegroundColor = GetColor(ConsoleColor.White);
            helpBackgroundColor = GetColor(_ConsoleColor.None);
            helpItemForegroundColor = GetColor(ConsoleColor.Green);
            helpItemBackgroundColor = GetColor(_ConsoleColor.None);
            legalForegroundColor = GetColor(ConsoleColor.Gray);
            legalBackgroundColor = GetColor(_ConsoleColor.None);
            resultForegroundColor = GetColor(_ConsoleColor.None);
            resultBackgroundColor = GetColor(_ConsoleColor.None);
            #endregion

            ///////////////////////////////////////////////////////////////////////////////////////////

            #region Section Colors
            announcementInfoForegroundColor = GetColor(ConsoleColor.White);
            announcementInfoBackgroundColor = GetColor(ConsoleColor.DarkRed);
            argumentInfoForegroundColor = GetColor(_ConsoleColor.None);
            argumentInfoBackgroundColor = GetColor(_ConsoleColor.None);
            callFrameInfoForegroundColor = GetColor(_ConsoleColor.None);
            callFrameInfoBackgroundColor = GetColor(_ConsoleColor.None);
            callStackInfoForegroundColor = GetColor(_ConsoleColor.None);
            callStackInfoBackgroundColor = GetColor(_ConsoleColor.None);
            complaintInfoForegroundColor = GetColor(_ConsoleColor.None);
            complaintInfoBackgroundColor = GetColor(_ConsoleColor.None);
            controlInfoForegroundColor = GetColor(_ConsoleColor.None);
            controlInfoBackgroundColor = GetColor(_ConsoleColor.None);
            customInfoForegroundColor = GetColor(_ConsoleColor.None);
            customInfoBackgroundColor = GetColor(_ConsoleColor.None);
            debuggerInfoForegroundColor = GetColor(_ConsoleColor.None);
            debuggerInfoBackgroundColor = GetColor(_ConsoleColor.None);
            engineInfoForegroundColor = GetColor(_ConsoleColor.None);
            engineInfoBackgroundColor = GetColor(_ConsoleColor.None);
            entityInfoForegroundColor = GetColor(_ConsoleColor.None);
            entityInfoBackgroundColor = GetColor(_ConsoleColor.None);
            flagInfoForegroundColor = GetColor(_ConsoleColor.None);
            flagInfoBackgroundColor = GetColor(_ConsoleColor.None);
            historyInfoForegroundColor = GetColor(_ConsoleColor.None);
            historyInfoBackgroundColor = GetColor(_ConsoleColor.None);
            hostInfoForegroundColor = GetColor(_ConsoleColor.None);
            hostInfoBackgroundColor = GetColor(_ConsoleColor.None);
            interpreterInfoForegroundColor = GetColor(_ConsoleColor.None);
            interpreterInfoBackgroundColor = GetColor(_ConsoleColor.None);
            objectInfoForegroundColor = GetColor(_ConsoleColor.None);
            objectInfoBackgroundColor = GetColor(_ConsoleColor.None);
            testInfoForegroundColor = GetColor(_ConsoleColor.None);
            testInfoBackgroundColor = GetColor(_ConsoleColor.None);
            tokenInfoForegroundColor = GetColor(_ConsoleColor.None);
            tokenInfoBackgroundColor = GetColor(_ConsoleColor.None);
            traceInfoForegroundColor = GetColor(_ConsoleColor.None);
            traceInfoBackgroundColor = GetColor(_ConsoleColor.None);
            variableInfoForegroundColor = GetColor(_ConsoleColor.None);
            variableInfoBackgroundColor = GetColor(_ConsoleColor.None);
            #endregion

            ///////////////////////////////////////////////////////////////////////////////////////////

            #region Return Code Colors
            okResultForegroundColor = GetColor(ConsoleColor.Green);
            okResultBackgroundColor = GetColor(_ConsoleColor.None);
            otherOkResultForegroundColor = GetColor(ConsoleColor.DarkGreen);
            otherOkResultBackgroundColor = GetColor(_ConsoleColor.None);
            otherErrorResultForegroundColor = GetColor(ConsoleColor.DarkRed);
            otherErrorResultBackgroundColor = GetColor(_ConsoleColor.None);
            errorResultForegroundColor = GetColor(ConsoleColor.Red);
            errorResultBackgroundColor = GetColor(_ConsoleColor.None);
            returnResultForegroundColor = GetColor(ConsoleColor.Blue);
            returnResultBackgroundColor = GetColor(_ConsoleColor.None);
            breakResultForegroundColor = GetColor(ConsoleColor.Yellow);
            breakResultBackgroundColor = GetColor(_ConsoleColor.None);
            continueResultForegroundColor = GetColor(ConsoleColor.Yellow);
            continueResultBackgroundColor = GetColor(_ConsoleColor.None);
            nullResultForegroundColor = GetColor(ConsoleColor.White);
            nullResultBackgroundColor = GetColor(_ConsoleColor.None);
            emptyResultForegroundColor = GetColor(ConsoleColor.White);
            emptyResultBackgroundColor = GetColor(_ConsoleColor.None);
            unknownResultForegroundColor = GetColor(ConsoleColor.Cyan);
            unknownResultBackgroundColor = GetColor(_ConsoleColor.None);
            #endregion

            ///////////////////////////////////////////////////////////////////////////////////////////

            #region Call Frame Colors
            nullCallFrameForegroundColor = GetColor(ConsoleColor.Gray);
            unknownCallFrameForegroundColor = GetColor(ConsoleColor.DarkGray);
            variableCallFrameForegroundColor = GetColor(ConsoleColor.White);
            uplevelCallFrameForegroundColor = GetColor(ConsoleColor.White);
            trackingCallFrameForegroundColor = GetColor(ConsoleColor.White);
            currentCallFrameForegroundColor = GetColor(ConsoleColor.White);
            procecureCallFrameForegroundColor = GetColor(ConsoleColor.Green);
            lambdaCallFrameForegroundColor = GetColor(ConsoleColor.Green);
            scopeCallFrameForegroundColor = GetColor(ConsoleColor.Green);
            aliasCallFrameForegroundColor = GetColor(ConsoleColor.White);
            globalCallFrameForegroundColor = GetColor(ConsoleColor.Yellow);
            linkedCallFrameForegroundColor = GetColor(ConsoleColor.DarkBlue);
            #endregion
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual void GetResultColors(
            ReturnCode code,
            Result result,
            ref ConsoleColor foregroundColor,
            ref ConsoleColor backgroundColor
            )
        {
            switch (code)
            {
                case ReturnCode.Ok:
                    {
                        foregroundColor = OkResultForegroundColor;
                        backgroundColor = OkResultBackgroundColor;
                        break;
                    }
                case ReturnCode.Error:
                    {
                        foregroundColor = ErrorResultForegroundColor;
                        backgroundColor = ErrorResultBackgroundColor;
                        break;
                    }
                case ReturnCode.Return:
                    {
                        foregroundColor = ReturnResultForegroundColor;
                        backgroundColor = ReturnResultBackgroundColor;
                        break;
                    }
                case ReturnCode.Break:
                    {
                        foregroundColor = BreakResultForegroundColor;
                        backgroundColor = BreakResultBackgroundColor;
                        break;
                    }
                case ReturnCode.Continue:
                    {
                        foregroundColor = ContinueResultForegroundColor;
                        backgroundColor = ContinueResultBackgroundColor;
                        break;
                    }
                default:
                    {
                        //
                        // NOTE: Check the return code (with "exceptions" enabled)
                        //       to see if it represents a success or failure.
                        //
                        if (RuntimeOps.IsSuccess(code, Exceptions))
                        {
                            //
                            // NOTE: This is an unknown "Ok" return code.
                            //
                            foregroundColor = OtherOkResultForegroundColor;
                            backgroundColor = OtherOkResultBackgroundColor;
                        }
                        else
                        {
                            //
                            // NOTE: This is an unknown "Error" return code.
                            //
                            foregroundColor = OtherErrorResultForegroundColor;
                            backgroundColor = OtherErrorResultBackgroundColor;
                        }
                        break;
                    }
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Formatting Methods
        #region Complaint Formatting Methods
        private bool HasComplaint(
            Interpreter interpreter
            )
        {
            string complaint = null;

            return HasComplaint(interpreter, ref complaint);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool HasComplaint(
            Interpreter interpreter,
            ref string complaint
            )
        {
            if (interpreter != null)
            {
                complaint = interpreter.Complaint;

                return !String.IsNullOrEmpty(complaint);
            }

            return false;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Result Formatting Methods
        private string FormatResult(
            string prefix,
            ReturnCode code,
            Result result,
            int errorLine,
            bool exceptions,
            bool ellipsis,
            bool replaceNewLines,
            bool strict
            )
        {
            ConsoleColor foregroundColor = ResultForegroundColor;
            ConsoleColor backgroundColor = ResultBackgroundColor;

            return FormatResult(prefix, code, result, errorLine, exceptions, ellipsis,
                replaceNewLines, strict, ref foregroundColor, ref backgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual string FormatResult(
            string prefix,
            ReturnCode code,
            Result result,
            int errorLine,
            bool exceptions,
            bool ellipsis,
            bool replaceNewLines,
            bool strict,
            ref ConsoleColor foregroundColor,
            ref ConsoleColor backgroundColor
            )
        {
            ReturnCode returnCode;

            if ((result == null) || (result.ReturnCode == ReturnCode.Ok))
                returnCode = code;
            else
                returnCode = result.ReturnCode;

            string formatted = FormatOps.Result(RuntimeOps.FormatResult(prefix,
                returnCode, result, errorLine, exceptions), ellipsis, replaceNewLines);

            if (!String.IsNullOrEmpty(formatted))
            {
                GetResultColors(returnCode, result, ref foregroundColor, ref backgroundColor);
            }
            else if (strict && (result == null))
            {
                foregroundColor = NullResultForegroundColor;
                backgroundColor = NullResultBackgroundColor;

                return FormatOps.DisplayNull;
            }
            else if (strict && (result.Length == 0))
            {
                foregroundColor = EmptyResultForegroundColor;
                backgroundColor = EmptyResultBackgroundColor;

                return FormatOps.DisplayEmpty;
            }
            else if (strict)
            {
                foregroundColor = UnknownResultForegroundColor;
                backgroundColor = UnknownResultBackgroundColor;

                return FormatOps.DisplayUnknown;
            }

            return formatted;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Introspection Methods
        #region Color Introspection Methods
        protected virtual List<PropertyInfo> GetColorProperties(
            Type type,
            MatchMode mode,
            string pattern,
            bool noCase,
            bool canRead,
            bool canWrite
            )
        {
            List<PropertyInfo> propertyList = new List<PropertyInfo>();

            if (type != null)
            {
                foreach (PropertyInfo propertyInfo in type.GetProperties(HostPropertyBindingFlags))
                {
                    if (propertyInfo != null)
                    {
                        if (propertyInfo.PropertyType == typeof(ConsoleColor))
                        {
                            if ((pattern == null) ||
                                StringOps.Match(mode, propertyInfo.Name, pattern, noCase))
                            {
                                if ((!canRead || propertyInfo.CanRead) &&
                                    (!canWrite || propertyInfo.CanWrite))
                                {
                                    propertyList.Add(propertyInfo);
                                }
                            }
                        }
                    }
                }
            }

            return propertyList;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Host Introspection Methods
        protected virtual string GetHostType(
            IHost host
            )
        {
            if (host != null)
            {
                Type type = host.GetType();

                if ((type == typeof(Console)) || (type == typeof(Core)) ||
                    (type == typeof(Default)) || (type == typeof(Diagnostic)) ||
                    (type == typeof(Null)))
                {
                    return String.Format(
                        FormatOps.DisplayFormat, host.GetType().Name.ToLower());
                }
                else
                {
                    return String.Format(
                        FormatOps.DisplayFormat, host.GetType().FullName);
                }
            }

            return FormatOps.DisplayNull;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual string GetHostInfo(
            IHost host,
            string name /* NOTE: The name of a property or method [with no arguments]. */
            )
        {
            //
            // HACK: Yes, this method is a nasty hack.  However, there is no
            //       other reasonable way to query all the host properties for
            //       display while not allowing exceptions to escape the
            //       WriteHostInfo method.
            //
            if (host != null)
            {
                try
                {
                    Type type = host.GetType();

                    do
                    {
                        object value;

                        try
                        {
                            value = type.InvokeMember(
                                name, MarshalOps.HostInfoBindingFlags,
                                null, host, null);
                        }
                        catch (Exception e)
                        {
                            type = type.BaseType;

                            if (type == typeof(object))
                                return FormatOps.DisplayException(e, true);

                            continue;
                        }

                        //
                        // NOTE: *SPECIAL* We want the display name of the
                        //       encoding, not the type name.
                        //
                        if (value is Encoding)
                            return ((Encoding)value).WebName;
                        else if (value != null)
                            return value.ToString();
                        else
                            return FormatOps.DisplayNull;
                    }
                    while (true);
                }
                catch (Exception e)
                {
                    return FormatOps.DisplayException(e, true);
                }
            }

            return FormatOps.DisplayNull;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Call Frame Introspection Methods
        protected virtual string GetCallFrameType(
            Interpreter interpreter,
            ICallFrame frame,
            bool special
            )
        {
            if (interpreter != null)
            {
                //
                // TODO: This is a very poor design.  Make this logic more
                //       customizable at runtime.
                //
                if (frame == null)
                    return NullCallFrameTypeName;
                else if (special && interpreter.IsVariableCallFrame(frame))
                    return VariableCallFrameTypeName;
                else if (interpreter.IsCurrentCallFrame(frame))
                    return CurrentCallFrameTypeName;
                else if (interpreter.IsProcedureCallFrame(frame))
                    return ProcedureCallFrameTypeName;
                else if (interpreter.IsGlobalCallFrame(frame))
                    return GlobalCallFrameTypeName;
                else if (CallFrameOps.IsLambda(frame))
                    return LambdaCallFrameTypeName;
                else if (CallFrameOps.IsScope(frame))
                    return ScopeCallFrameTypeName;
                else if (CallFrameOps.IsAlias(frame))
                    return AliasCallFrameTypeName;
                else if (CallFrameOps.IsUplevel(frame))
                    return UplevelCallFrameTypeName;
                else if (CallFrameOps.IsTracking(frame))
                    return TrackingCallFrameTypeName;
                else if (interpreter.IsCallFrameInCallStack(frame))
                    //
                    // NOTE: Yes, this normally returns only whitespace.
                    //
                    return NormalCallFrameTypeName;
            }

            return UnknownCallFrameTypeName;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual ConsoleColor GetCallFrameColor(
            Interpreter interpreter,
            ICallFrame frame,
            bool special
            )
        {
            if (interpreter != null)
            {
                //
                // TODO: This is a very poor design.  Make this logic more
                //       customizable at runtime.
                //
                if (frame == null)
                    return NullCallFrameForegroundColor;
                else if (special && interpreter.IsVariableCallFrame(frame))
                    return VariableCallFrameForegroundColor;
                else if (interpreter.IsCurrentCallFrame(frame))
                    return CurrentCallFrameForegroundColor;
                else if (interpreter.IsProcedureCallFrame(frame))
                    return ProcecureCallFrameForegroundColor;
                else if (interpreter.IsGlobalCallFrame(frame))
                    return GlobalCallFrameForegroundColor;
                else if (CallFrameOps.IsLambda(frame))
                    return LambdaCallFrameForegroundColor;
                else if (CallFrameOps.IsScope(frame))
                    return ScopeCallFrameForegroundColor;
                else if (CallFrameOps.IsAlias(frame))
                    return AliasCallFrameForegroundColor;
                else if (CallFrameOps.IsUplevel(frame))
                    return UplevelCallFrameForegroundColor;
                else if (CallFrameOps.IsTracking(frame))
                    return TrackingCallFrameForegroundColor;
                else if (interpreter.IsCallFrameInCallStack(frame))
                    return CallFrameInfoForegroundColor;
            }

            return UnknownCallFrameForegroundColor;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool BuildCallFrameInfoList(
            Interpreter interpreter,
            ICallFrame frame,
            bool linked,
            bool special,
            bool variables,
            bool empty, /* NOT USED */
            ref StringPairList list
            )
        {
            if (list == null)
                list = new StringPairList();

            if (list.Count == 0)
            {
                list.Add("Frame");
                list.Add((IPair<string>)null);
            }

            if (frame != null)
            {
                list.Add(frame.ToList());
                list.Add("special", special.ToString());
                list.Add((IPair<string>)null);
                list.Add("Links");
                list.Add((IPair<string>)null);
                list.Add("inbound", linked.ToString());
                list.Add("outbound", (frame.Link != null).ToString());
                list.Add((IPair<string>)null);

                if (interpreter != null)
                {
                    list.Add("inCallStack", interpreter.IsCallFrameInCallStack(frame).ToString());
                    list.Add("isGlobal", interpreter.IsGlobalCallFrame(frame).ToString());
                    list.Add("isCurrent", interpreter.IsCurrentCallFrame(frame).ToString());
                    list.Add("isProcedure", interpreter.IsProcedureCallFrame(frame).ToString());
                    list.Add("isVariable", interpreter.IsVariableCallFrame(frame).ToString());
                }

                list.Add("typeName", GetCallFrameType(interpreter, frame, special));
                list.Add("color", GetCallFrameColor(interpreter, frame, special).ToString());

                if (variables)
                {
                    VariableDictionary frameVariables = frame.Variables;

                    if ((frameVariables != null) && (frameVariables.Count > 0))
                    {
                        list.Add((IPair<string>)null);
                        list.Add("Variables");
                        list.Add((IPair<string>)null);
                        list.Add(frameVariables.ToString());
                    }
                }
            }
            else
            {
                list.Add(FormatOps.DisplayNull);
            }

            return true;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Object Introspection Methods
        private static IObject GetObjectFromValue(
            Interpreter interpreter,
            object value
            )
        {
            string name = RuntimeOps.GetStringFromObject(value);

            if (name != null)
            {
                IObject @object = null;

                if ((interpreter != null) &&
                    interpreter.GetObject(name, true, true, ref @object) == ReturnCode.Ok)
                {
                    return @object;
                }
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static Type GetWrappedObjectType(
            IObject @object
            )
        {
            _Wrappers._Object objectWrapper = @object as _Wrappers._Object;

            if ((objectWrapper != null) && (objectWrapper.@object != null))
                return objectWrapper.@object.GetType();

            return null;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool BuildObjectInfoList(
            Interpreter interpreter,
            IObject @object,
            bool empty, /* NOT USED */
            ref StringPairList list
            )
        {
            if (list == null)
                list = new StringPairList();

            if (list.Count == 0)
            {
                list.Add("Object");
                list.Add((IPair<string>)null);
            }

            if (@object != null)
            {
                list.Add("ReferenceCount", @object.ReferenceCount.ToString());
                list.Add("ObjectFlags", @object.Flags.ToString());

                //
                // NOTE: What type is the wrapper for this object?
                //
                Type objectType = GetWrappedObjectType(@object);

                if (objectType != null)
                {
                    if (Object.ReferenceEquals(objectType.Assembly, GlobalState.GetAssembly()))
                        list.Add("ObjectType", objectType.FullName);
                    else
                        list.Add("ObjectType", objectType.AssemblyQualifiedName);
                }

                if (@object.Value != null)
                {
                    //
                    // NOTE: What type is this object?
                    //
                    Type valueType = @object.Value.GetType();

                    if (Object.ReferenceEquals(valueType.Assembly, GlobalState.GetAssembly()))
                        list.Add("ValueType", valueType.FullName);
                    else
                        list.Add("ValueType", valueType.AssemblyQualifiedName);

                    list.Add("ToString", @object.Value.ToString());
                }
                else
                {
                    list.Add("Value", FormatOps.DisplayNull);
                }
            }
            else
            {
                list.Add(FormatOps.DisplayNull);
            }

            return true;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Variable Introspection Methods
        protected internal virtual bool BuildVariableInfoList(
            Interpreter interpreter,
            IVariable variable,
            bool elements, /* show all array elements? */
            bool links, /* show links? */
            bool empty,
            ref StringPairList list
            )
        {
            BuildVariableInfoList(interpreter, variable, elements, empty, ref list);

            if (links)
            {
                IVariable link = variable.Link;
                int count = 0;

                while (link != null)
                {
                    list.Add((IPair<string>)null);
                    list.Add(String.Format("Link #{0}", count));
                    list.Add((IPair<string>)null);

                    BuildVariableInfoList(interpreter, link, elements, empty, ref list);

                    count++;
                    link = link.Link;
                }
            }

            return true;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool BuildVariableInfoList(
            Interpreter interpreter,
            IVariable variable,
            bool elements, /* show all array elements? */
            bool empty, /* NOT USED */
            ref StringPairList list
            )
        {
            if (list == null)
                list = new StringPairList();

            if (list.Count == 0)
            {
                list.Add("Variable");
                list.Add((IPair<string>)null);
            }

            if (variable != null)
            {
                list.Add("Flags", variable.Flags.ToString());

                string name = (variable.Name != null) ?
                    variable.Name : FormatOps.DisplayNull;

                list.Add("Name", name);

                if (EntityOps.IsArray2(variable))
                {
                    ObjectDictionary arrayValue = variable.ArrayValue;

                    if (arrayValue != null)
                    {
                        list.Add("ArrayValue",
                            arrayValue.Count.ToString());

                        if (elements && (arrayValue.Count > 0))
                        {
                            list.Add((IPair<string>)null);

                            foreach (KeyValuePair<string, object> pair in arrayValue)
                            {
                                string value = (pair.Value != null) ?
                                    RuntimeOps.GetStringFromObject(pair.Value) : FormatOps.DisplayNull;

                                list.Add(new StringPair(pair.Key, value));

                                IObject @object = GetObjectFromValue(interpreter, pair.Value);

                                if (@object != null)
                                    BuildObjectInfoList(interpreter, @object, empty, ref list);
                            }
                        }

                        //
                        // NOTE: If we have a valid interpreter context, check
                        //       for array searches belonging to this variable.
                        //
                        if (interpreter != null)
                        {
                            lock (interpreter.SyncRoot)
                            {
                                StringPairList arraySearches = new StringPairList();

                                if (interpreter.ArraySearches != null)
                                {
                                    foreach (KeyValuePair<string, ArraySearch> pair
                                            in interpreter.ArraySearches)
                                    {
                                        if ((pair.Value != null) &&
                                            Object.ReferenceEquals(pair.Value.Variable, variable))
                                        {
                                            arraySearches.Add(String.Format(
                                                "Name #{0}", arraySearches.Count), pair.Key);
                                        }
                                    }
                                }

                                list.Add((IPair<string>)null);
                                list.Add("ArraySearches",
                                    arraySearches.Count.ToString());

                                if (arraySearches.Count > 0)
                                {
                                    list.Add((IPair<string>)null);
                                    list.Add(arraySearches);
                                }
                            }
                        }
                    }
                    else
                    {
                        list.Add("ArrayValue", FormatOps.DisplayNull);
                    }
                }
                else
                {
                    string value = (variable.Value != null) ?
                        RuntimeOps.GetStringFromObject(variable.Value) : FormatOps.DisplayNull;

                    list.Add("Value", value);

                    IObject @object = GetObjectFromValue(interpreter, variable.Value);

                    if (@object != null)
                    {
                        list.Add((IPair<string>)null);
                        list.Add("Object");
                        list.Add((IPair<string>)null);

                        BuildObjectInfoList(interpreter, @object, empty, ref list);
                    }
                }

                if (variable.Frame != null)
                {
                    list.Add((IPair<string>)null);
                    list.Add("Frame");
                    list.Add((IPair<string>)null);

                    ICallFrame link = variable.Frame;
                    int count = 0;

                    while (link != null)
                    {
                        list.Add(String.Format(
                            "Name #{0}", count), link.Name);

                        list.Add(String.Format(
                            "Flags #{0}", count), link.Flags.ToString());

                        count++;
                        link = link.Link;
                    }
                }

                if (variable.Traces != null)
                {
                    list.Add((IPair<string>)null);
                    list.Add("Traces",
                        variable.Traces.Count.ToString());

                    if (variable.Traces.Count > 0)
                    {
                        list.Add((IPair<string>)null);

                        foreach (ITrace trace in variable.Traces)
                        {
                            if (trace != null)
                            {
                                //
                                // NOTE: Is this an interpreter-wide trace?
                                //
                                bool global = FlagOps.HasFlags(
                                    trace.TraceFlags, TraceFlags.Global, true);

#if ISOLATED_PLUGINS
                                //
                                // NOTE: Get the plugin for this trace.
                                //
                                IPlugin plugin = trace.Plugin;

                                if (!RuntimeOps.IsPluginIsolated(plugin))
#endif
                                {
                                    TraceCallback traceCallback = trace.Callback;

                                    if ((traceCallback != null) &&
                                        (traceCallback.Method != null))
                                    {
                                        list.Add(global ? "Interpreter" : "Variable",
                                            FormatOps.DelegateMethodName(
                                                traceCallback.Method.DeclaringType,
                                                traceCallback.Method.Name, false));
                                    }
                                }
#if ISOLATED_PLUGINS
                                else

                                {
                                    list.Add(global ? "Interpreter" : "Variable",
                                        FormatOps.DelegateMethodName(
                                            trace.TypeName, trace.MethodName));
                                }
#endif
                            }
                        }
                    }
                }
            }
            else
            {
                list.Add(FormatOps.DisplayNull);
            }

            return true;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Test Methods
        protected virtual bool InTestMode()
        {
            return FlagOps.HasFlags(GetFlags(), HostFlags.Test, true);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool HasTestFlags(HostTestFlags hasFlags, bool all)
        {
            return FlagOps.HasFlags(GetTestFlags(), hasFlags, all);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Core Write Helper Methods
        protected virtual bool WriteCore(
            WriteCharCallback writeCharCallback,
            WriteLineCallback writeLineCallback,
            char value,
            int count,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            if ((writeCharCallback != null) &&
                (!newLine || (writeLineCallback != null)) &&
                (count >= 0))
            {
                try
                {
                    bool color = DoesSupportColor();

                    ConsoleColor savedForegroundColor = _ConsoleColor.None;
                    ConsoleColor savedBackgroundColor = _ConsoleColor.None;

                    if (color && !GetColors(ref savedForegroundColor, ref savedBackgroundColor))
                        return false;

                    try
                    {
                        if (color && !SetColors(true, true, foregroundColor, backgroundColor))
                            return false;

                        while (count-- > 0)
                            writeCharCallback(value); /* throw */

                        if (newLine)
                            writeLineCallback(); /* throw */

                        return true;
                    }
                    finally
                    {
                        if (color)
                            /* IGNORED */
                            SetColors(true, true, savedForegroundColor, savedBackgroundColor);
                    }
                }
                catch
                {
                    // do nothing.
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool WriteCore(
            WriteStringCallback writeStringCallback,
            WriteLineCallback writeLineCallback,
            string value,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            if ((writeStringCallback != null) &&
                (!newLine || (writeLineCallback != null)))
            {
                try
                {
                    bool color = DoesSupportColor();

                    ConsoleColor savedForegroundColor = _ConsoleColor.None;
                    ConsoleColor savedBackgroundColor = _ConsoleColor.None;

                    if (color && !GetColors(ref savedForegroundColor, ref savedBackgroundColor))
                        return false;

                    try
                    {
                        if (color && !SetColors(true, true, foregroundColor, backgroundColor))
                            return false;

                        writeStringCallback(value); /* throw */

                        if (newLine)
                            writeLineCallback(); /* throw */

                        return true;
                    }
                    finally
                    {
                        if (color)
                            /* IGNORED */
                            SetColors(true, true, savedForegroundColor, savedBackgroundColor);
                    }
                }
                catch
                {
                    // do nothing.
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool WriteCore(
            HostWriteType hostWriteType,
            char value,
            int count,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            if (count >= 0)
            {
                try
                {
                    bool color = DoesSupportColor();

                    ConsoleColor savedForegroundColor = _ConsoleColor.None;
                    ConsoleColor savedBackgroundColor = _ConsoleColor.None;

                    if (color && !GetColors(ref savedForegroundColor, ref savedBackgroundColor))
                        return false;

                    try
                    {
                        if (color && !SetColors(true, true, foregroundColor, backgroundColor))
                            return false;

                        switch (hostWriteType)
                        {
                            case HostWriteType.Normal:
                                {
                                    while (count-- > 0)
                                        if (!Write(value))
                                            return false;

                                    if (newLine)
                                        if (!WriteLine())
                                            return false;

                                    return true;
                                }
                            case HostWriteType.Debug:
                                {
                                    while (count-- > 0)
                                        if (!WriteDebug(value))
                                            return false;

                                    if (newLine)
                                        if (!WriteDebugLine())
                                            return false;

                                    return true;
                                }
                            case HostWriteType.Error:
                                {
                                    while (count-- > 0)
                                        if (!WriteError(value))
                                            return false;

                                    if (newLine)
                                        if (!WriteErrorLine())
                                            return false;

                                    return true;
                                }
                        }
                    }
                    finally
                    {
                        if (color)
                            /* IGNORED */
                            SetColors(true, true, savedForegroundColor, savedBackgroundColor);
                    }
                }
                catch
                {
                    // do nothing.
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool WriteCore(
            HostWriteType hostWriteType,
            string value,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            try
            {
                bool color = DoesSupportColor();

                ConsoleColor savedForegroundColor = _ConsoleColor.None;
                ConsoleColor savedBackgroundColor = _ConsoleColor.None;

                if (color && !GetColors(ref savedForegroundColor, ref savedBackgroundColor))
                    return false;

                try
                {
                    if (color && !SetColors(true, true, foregroundColor, backgroundColor))
                        return false;

                    switch (hostWriteType)
                    {
                        case HostWriteType.Normal:
                            {
                                return newLine ? WriteLine(value) : Write(value);
                            }
                        case HostWriteType.Debug:
                            {
                                return newLine ? WriteDebugLine(value) : WriteDebug(value);
                            }
                        case HostWriteType.Error:
                            {
                                return newLine ? WriteErrorLine(value) : WriteError(value);
                            }
                    }
                }
                finally
                {
                    if (color)
                        /* IGNORED */
                        SetColors(true, true, savedForegroundColor, savedBackgroundColor);
                }
            }
            catch
            {
                // do nothing.
            }

            return false;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Content Section Methods
        protected virtual bool DoesSupportColor()
        {
            return FlagOps.HasFlags(GetFlags(), HostFlags.AllColors, false);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool DoesSupportSizing()
        {
            return FlagOps.HasFlags(GetFlags(), HostFlags.Sizing, true);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool DoesSupportPositioning()
        {
            return FlagOps.HasFlags(GetFlags(), HostFlags.Positioning, true);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool IsAtLeastSize(
            HostFlags flags,
            HostFlags hasFlags
            )
        {
            if (!FlagOps.HasFlags(hasFlags, HostFlags.AllSizes, false))
                return false;

            if (FlagOps.HasFlags(flags, HostFlags.UnlimitedSize, true))
                return true;

            if (FlagOps.HasFlags(hasFlags, HostFlags.SuperJumboSize, true) &&
                FlagOps.HasFlags(flags, HostFlags.SuperJumboSize, true))
            {
                return true;
            }

            if (FlagOps.HasFlags(hasFlags, HostFlags.JumboSize, true) &&
                (FlagOps.HasFlags(flags, HostFlags.SuperJumboSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.JumboSize, true)))
            {
                return true;
            }

            if (FlagOps.HasFlags(hasFlags, HostFlags.SuperFullSize, true) &&
                (FlagOps.HasFlags(flags, HostFlags.SuperJumboSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.JumboSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.SuperFullSize, true)))
            {
                return true;
            }

            if (FlagOps.HasFlags(hasFlags, HostFlags.FullSize, true) &&
                (FlagOps.HasFlags(flags, HostFlags.SuperJumboSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.JumboSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.SuperFullSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.FullSize, true)))
            {
                return true;
            }

            if (FlagOps.HasFlags(hasFlags, HostFlags.CompactSize, true) &&
                (FlagOps.HasFlags(flags, HostFlags.SuperJumboSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.JumboSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.SuperFullSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.FullSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.CompactSize, true)))
            {
                return true;
            }

            if (FlagOps.HasFlags(hasFlags, HostFlags.MinimumSize, true) &&
                (FlagOps.HasFlags(flags, HostFlags.SuperJumboSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.JumboSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.SuperFullSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.FullSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.CompactSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.MinimumSize, true)))
            {
                return true;
            }

            if (FlagOps.HasFlags(hasFlags, HostFlags.ZeroSize, true) &&
                (FlagOps.HasFlags(flags, HostFlags.SuperJumboSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.JumboSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.SuperFullSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.FullSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.CompactSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.MinimumSize, true) ||
                 FlagOps.HasFlags(flags, HostFlags.ZeroSize, true)))
            {
                return true;
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool DoesHeaderFit(
            HeaderFlags headerFlags,
            HostFlags hostFlags
            )
        {
            HostFlags hasFlags;

            if ((sectionSizes != null) &&
                sectionSizes.TryGetValue(headerFlags, out hasFlags) &&
                IsAtLeastSize(hostFlags, hasFlags))
            {
                return true;
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool BeginSection(
            bool nextRow,
            int minimumLeft,
            ref int maximumTop,
            ref int savedTop,
            ref int count
            )
        {
            if (OutputStyle == OutputStyle.Boxed)
            {
                if (nextRow || (savedTop == _Position.Invalid))
                {
                    //
                    // NOTE: Advance to the start of the next line.
                    //
                    hostLeft = minimumLeft;
                    hostTop = ++maximumTop;

                    savedTop = hostTop;
                    count = 0;
                }
                else
                {
                    hostLeft++;
                    hostTop = savedTop;

                    count++;
                }

                return true;
            }
            else if (OutputStyle == OutputStyle.Formatted)
            {
                //
                // NOTE: Advance to the start of the next line.
                //
                hostLeft = minimumLeft;
                hostTop = ++maximumTop;

                //
                // NOTE: For formatted mode, always update the
                //       actual output position with our internal
                //       variables.
                //
                if (DoesSupportPositioning() &&
                    !SetPosition(hostLeft, hostTop))
                {
                    return false;
                }

                return true;
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool EndSection(
            ref int maximumLeft,
            ref int maximumTop
            )
        {
            //
            // NOTE: For formatted mode, always update the
            //       internal variables to our actual output
            //       position.
            //
            if ((OutputStyle == OutputStyle.Formatted) &&
                DoesSupportPositioning() &&
                !GetPosition(ref hostLeft, ref hostTop))
            {
                return false;
            }

            //
            // NOTE: For both supported modes, check and update
            //       the maximum output positions we have seen
            //       so far.
            //
            if ((OutputStyle == OutputStyle.Formatted) ||
                (OutputStyle == OutputStyle.Boxed))
            {
                if (hostLeft > maximumLeft)
                    maximumLeft = hostLeft;

                if (hostTop > maximumTop)
                    maximumTop = hostTop;

                return true;
            }

            return false;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IIdentifier Members
        private IdentifierKind kind;
        public virtual IdentifierKind Kind
        {
            get { CheckDisposed(); return kind; }
            set { CheckDisposed(); kind = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private Guid id;
        public virtual Guid Id
        {
            get { CheckDisposed(); return id; }
            set { CheckDisposed(); id = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string name;
        public virtual string Name
        {
            get { CheckDisposed(); return name; }
            set { CheckDisposed(); name = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string group;
        public virtual string Group
        {
            get { CheckDisposed(); return group; }
            set { CheckDisposed(); group = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string description;
        public virtual string Description
        {
            get { CheckDisposed(); return description; }
            set { CheckDisposed(); description = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private IClientData clientData;
        public virtual IClientData ClientData
        {
            get { CheckDisposed(); return clientData; }
            set { CheckDisposed(); clientData = value; }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IInteractiveHost Members
        public virtual ReturnCode BeginProcessing(
            int levels,
            ref string text,
            ref Result error
            )
        {
            CheckDisposed();

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual ReturnCode EndProcessing(
            int levels,
            ref string text,
            ref Result error
            )
        {
            CheckDisposed();

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual ReturnCode DoneProcessing(
            int levels,
            ref Result error
            )
        {
            CheckDisposed();

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string title;
        public virtual string Title
        {
            get { CheckDisposed(); return title; }
            set { CheckDisposed(); title = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool IsInputRedirected(); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract ReturnCode Prompt(
            PromptType type,
            ref PromptFlags flags,
            ref Result error
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool IsOpen(); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool Pause(); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private HostFlags hostFlags = HostFlags.Invalid;
        public virtual HostFlags GetFlags()
        {
            CheckDisposed();

            if (hostFlags == HostFlags.Invalid)
            {
                //
                // NOTE: We support the CanExit, CanForceExit,
                //       and Exiting properties.
                //
                hostFlags = HostFlags.Exit;
            }

            return hostFlags;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool ReadLine(
            ref string value
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool Write(
            char value
            )
        {
            CheckDisposed();

            return Write(value, false);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool Write(
            string value
            )
        {
            CheckDisposed();

            return Write(value, false);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool WriteLine(); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteLine(
            string value
            )
        {
            CheckDisposed();

            return Write(value, true);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteResultLine(
            ReturnCode code,
            Result result
            )
        {
            CheckDisposed();

            return WriteResult(code, result, true);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteResultLine(
            ReturnCode code,
            Result result,
            int errorLine
            )
        {
            CheckDisposed();

            return WriteResult(code, result, errorLine, true);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IEngineHost Members
        private bool canExit;
        public virtual bool CanExit
        {
            get { CheckDisposed(); return canExit; }
            set { CheckDisposed(); canExit = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool canForceExit;
        public virtual bool CanForceExit
        {
            get { CheckDisposed(); return canForceExit; }
            set { CheckDisposed(); canForceExit = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool exiting;
        public virtual bool Exiting
        {
            get { CheckDisposed(); return exiting; }
            set { CheckDisposed(); exiting = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract ReturnCode CreateThread(
            ThreadStart start,
            int maxStackSize,
            bool userInterface,
            bool isBackground,
            ref Thread thread,
            ref Result error
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract ReturnCode CreateThread(
            ParameterizedThreadStart start,
            int maxStackSize,
            bool userInterface,
            bool isBackground,
            ref Thread thread,
            ref Result error
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract ReturnCode QueueWorkItem(
            WaitCallback callback,
            object state,
            ref Result error
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract ReturnCode GetStream(
            string path,
            FileMode mode,
            FileAccess access,
            FileShare share,
            int bufferSize,
            FileOptions options,
            bool resolve,
            ref string fullPath,
            ref Stream stream,
            ref Result error
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract ReturnCode GetScript(
            string name,
            ref ScriptFlags flags,
            ref Result result
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool Sleep(
            int milliseconds
            ); /* PRIMITIVE */
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IStreamHost Members
        public virtual Stream DefaultIn
        {
            get { CheckDisposed(); return In; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual Stream DefaultOut
        {
            get { CheckDisposed(); return Out; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual Stream DefaultError
        {
            get { CheckDisposed(); return Error; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract Stream In { get; set; } /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract Stream Out { get; set; } /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract Stream Error { get; set; } /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract Encoding InputEncoding { get; set; } /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract Encoding OutputEncoding { get; set; } /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract Encoding ErrorEncoding { get; set; } /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool ResetIn(); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool ResetOut(); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool ResetError(); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool IsOutputRedirected(); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool IsErrorRedirected(); /* PRIMITIVE */
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IDebugHost Members
        public abstract IHost Clone(); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract IHost Clone(
            Interpreter interpreter
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract HostTestFlags GetTestFlags(); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract ReturnCode Cancel(
            ref Result error
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract ReturnCode Exit(
            ref Result error
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool WriteDebugLine(); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteDebugLine(
            string value
            )
        {
            CheckDisposed();

            return WriteDebug(value, true);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteDebug(
            char value
            )
        {
            CheckDisposed();

            return WriteDebug(value, false);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool WriteDebug(
            char value,
            bool newLine
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteDebug(
            string value
            )
        {
            CheckDisposed();

            return WriteDebug(value, false);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool WriteDebug(
            string value,
            bool newLine
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteDebug(
            string value,
            bool newLine,
            ConsoleColor foregroundColor
            )
        {
            CheckDisposed();

            return WriteDebug(value, newLine, foregroundColor, DebugBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteDebug(
            string value,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            return WriteCore(HostWriteType.Debug, value, newLine,
                foregroundColor, backgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool WriteErrorLine(); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteErrorLine(
            string value
            )
        {
            CheckDisposed();

            return WriteError(value, true);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteError(
            char value
            )
        {
            CheckDisposed();

            return WriteError(value, false);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool WriteError(
            char value,
            bool newLine
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteError(
            string value
            )
        {
            CheckDisposed();

            return WriteError(value, false);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool WriteError(
            string value,
            bool newLine
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteError(
            string value,
            bool newLine,
            ConsoleColor foregroundColor
            )
        {
            CheckDisposed();

            return WriteError(value, newLine, foregroundColor, ErrorBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteError(
            string value,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            return WriteCore(HostWriteType.Error, value, newLine,
                foregroundColor, backgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteResult(
            ReturnCode code,
            Result result,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteResult(code, result, false, newLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteResult(
            ReturnCode code,
            Result result,
            bool raw,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteResult(null, code, result, 0, raw, newLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteResult(
            ReturnCode code,
            Result result,
            int errorLine,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteResult(null, code, result, errorLine, false, newLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteResult(
            ReturnCode code,
            Result result,
            int errorLine,
            bool raw,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteResult(null, code, result, errorLine, raw, newLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteResult(
            string prefix,
            ReturnCode code,
            Result result,
            int errorLine,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteResult(prefix, code, result, errorLine, false, newLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteResult(
            string prefix,
            ReturnCode code,
            Result result,
            int errorLine,
            bool raw,
            bool newLine
            )
        {
            CheckDisposed();

            ReturnCode returnCode;

            if ((result == null) || (result.ReturnCode == ReturnCode.Ok))
                returnCode = code;
            else
                returnCode = result.ReturnCode;

            bool wrote;
            string formatted;

            if (raw)
                formatted = result;
            else
                //
                // TODO: Possibly have the caller pass in the exceptions argument
                //       and the other three arguments that are simply hard-coded
                //       to false here (i.e. ellipsis, replaceNewLines, and strict).
                //
                formatted = FormatResult(prefix, returnCode, result, errorLine,
                    Exceptions, false, false, false);

            if (!String.IsNullOrEmpty(formatted))
            {
                ConsoleColor foregroundColor = ResultForegroundColor;
                ConsoleColor backgroundColor = ResultBackgroundColor;

                GetResultColors(returnCode, result, ref foregroundColor, ref backgroundColor);

                wrote = Write(formatted, newLine, foregroundColor, backgroundColor);
            }
            else
            {
                wrote = false;
            }

            ///////////////////////////////////////////////////////////////////

            #region Break-On-Exiting Code
            //
            // NOTE: This block of code was primarily designed to catch
            //       termination and/or resource disposal errors encountered
            //       during the disposal of the Interpreter class (e.g. those
            //       raised by plugins).  Originally, this code was added while
            //       tracking down a particular interpreter disposal issue and
            //       it was intended to be temporary; however, it ended up
            //       being far too useful to remove.
            //
#if BREAK_ON_EXITING
            //
            // NOTE: If we (the host) is exiting, complain loudly about non-Ok
            //       return codes.
            //
            if (exiting)
            {
                //
                // NOTE: Prevent infinite recursion because the DebugOps methods
                //       call this function.
                //
                exiting = false;

                if (returnCode != ReturnCode.Ok)
                {
                    //
                    // NOTE: Is a debugger attached to this process?
                    //
                    if (DebugOps.IsAttached())
                    {
                        //
                        // NOTE: Break into the attached debugger.
                        //
                        DebugOps.Break(null, this, DebugOps.GetMethod(0),
                            false);
                    }
                    else
                    {
                        //
                        // NOTE: Otherwise, just complain loudly.
                        //
                        string message = String.Format(
                            "received failure return code {0} while exiting",
                            returnCode);

                        DebugOps.Fail(null, this, DebugOps.GetMethod(0),
                            message, result, false);
                    }
                }

                exiting = true;
            }
#endif
            #endregion

            ///////////////////////////////////////////////////////////////////

            return wrote;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IInformationHost Members
        public virtual bool SavePosition()
        {
            CheckDisposed();

            if (OutputStyle != OutputStyle.Boxed)
                return true;

            if (!DoesSupportPositioning())
                return true;

            return GetPosition(ref hostLeft, ref hostTop);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool RestorePosition(bool newLine)
        {
            CheckDisposed();

            if (OutputStyle != OutputStyle.Boxed)
                return true;

            if (!DoesSupportPositioning())
                return !newLine || WriteLine();

            if (newLine)
                return SetPosition(0, hostTop);
            else
                return SetPosition(hostLeft, hostTop);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteAnnouncementInfo(
            Interpreter interpreter,
            string value,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteAnnouncementInfo(interpreter, value, newLine,
                AnnouncementInfoForegroundColor, AnnouncementInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteAnnouncementInfo(
            Interpreter interpreter,
            string value,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            if (!Write(Environment.NewLine +
                    StringOps.PadCenter(value, ContentWidth - ContentMargin, Characters.Space),
                    newLine, foregroundColor, backgroundColor))
            {
                return false;
            }

            return true;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteArgumentInfo(
            Interpreter interpreter,
            ReturnCode code,
            BreakpointType breakpointType,
            string breakpointName,
            ArgumentList arguments,
            Result result,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteArgumentInfo(interpreter, code, breakpointType, breakpointName, arguments,
                result, empty, newLine, ArgumentInfoForegroundColor, ArgumentInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteArgumentInfo(
            Interpreter interpreter,
            ReturnCode code,
            BreakpointType breakpointType,
            string breakpointName,
            ArgumentList arguments,
            Result result,
            bool empty, /* NOT USED */
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            StringPairList list = new StringPairList();

            list.Add("ArgumentInfo");
            list.Add((IPair<string>)null);

            list.Add("BreakpointType", breakpointType.ToString());
            list.Add("BreakpointName", FormatOps.DisplayString(breakpointName));

            if (arguments != null)
            {
                list.Add((IPair<string>)null);
                list.Add("Arguments",
                    arguments.ToString(ToStringFlags.NameAndValue, null, false));
            }

            list.Add((IPair<string>)null);
            list.Add("Code", code.ToString());
            list.Add("Result", FormatOps.DisplayResult(result, true, true));

            if (OutputStyle == OutputStyle.Boxed)
                return WriteBox(ArgumentInfoBoxName, list, null, false, true,
                    ref hostLeft, ref hostTop, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.Formatted)
                return Write(list, newLine, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.None)
                return true;
            else
                return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteCallFrame(
            Interpreter interpreter,
            ICallFrame frame,
            string suffix,
            char separator,
            bool linked,
            bool special,
            bool variables,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            string value = (frame != null) ? ": " + frame.ToString(true) : String.Empty;

            string prefix = linked ? String.Empty : String.Empty; /* TODO: Why? */

            string type = linked ?
                LinkedCallFrameTypeName : GetCallFrameType(interpreter, frame, special);

            ConsoleColor foregroundColor = linked ?
                LinkedCallFrameForegroundColor : GetCallFrameColor(interpreter, frame, special);

            string formatted = Characters.OpenBracket + String.Format("{0}{1}",
                prefix, type) + value + Characters.CloseBracket;

            if (!String.IsNullOrEmpty(suffix)) formatted += suffix;

            int separatorLength = Math.Min(formatted.Length, ContentWidth - ContentMargin);

            return Write(formatted, newLine, foregroundColor, CallFrameInfoBackgroundColor) &&
                (separator != NoSeparator) ?
                    Write(VariableCallFrameSeparator, separatorLength, true) : true;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteCallFrameInfo(
            Interpreter interpreter,
            ICallFrame frame,
            bool linked,
            bool special,
            bool variables,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteCallFrameInfo(
                interpreter, frame, linked, special, variables, empty, newLine,
                CallFrameInfoForegroundColor, CallFrameInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteCallFrameInfo(
            Interpreter interpreter,
            ICallFrame frame,
            bool linked,
            bool special,
            bool variables,
            bool empty,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            if (frame != null)
            {
                StringPairList list = null;

                if (BuildCallFrameInfoList(
                        interpreter, frame, linked, special, variables, empty, ref list))
                {
                    if (OutputStyle == OutputStyle.Boxed)
                        return WriteBox(CallFrameInfoBoxName, list, null, false, true,
                            ref hostLeft, ref hostTop, foregroundColor, backgroundColor);
                    else if (OutputStyle == OutputStyle.Formatted)
                        return Write(list, newLine, foregroundColor, backgroundColor);
                    else if (OutputStyle == OutputStyle.None)
                        return true;
                    else
                        return false;
                }
                else
                {
                    return false;
                }
            }
            else
            {
                return true;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteCallStack(
            Interpreter interpreter,
            CallStack callStack,
            bool all,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteCallStack(interpreter, callStack, CallStackLimit, all, empty, newLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteCallStack(
            Interpreter interpreter,
            CallStack callStack,
            int limit,
            bool all,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            if ((interpreter != null) && (callStack != null))
            {
                lock (interpreter.SyncRoot) /* TRANSACTIONAL */
                {
                    ICallFrame variableFrame = null;

                    if (all)
                    {
                        variableFrame = interpreter.GetVariableFrame(
                            null, VariableFlags.None);

                        if (variableFrame != null)
                        {
                            if (!WriteCallFrame(
                                    interpreter, variableFrame,
                                    VariableCallFrameSuffix,
                                    VariableCallFrameSeparator,
                                    false, true, false, empty, true))
                            {
                                return false;
                            }

                            if ((variableFrame.Link != null) &&
                                !WriteCallFrame(
                                    interpreter, variableFrame.Link,
                                    null, NoSeparator,
                                    true, true, false, empty, true))
                            {
                                return false;
                            }
                        }
                    }

                    CallStack newCallStack = null;
                    Result error = null;
                    int count = 0;

                    if (CallFrameOps.Traverse(
                            interpreter, callStack, variableFrame,
                            limit, all, ref newCallStack,
                            ref error) == ReturnCode.Ok)
                    {
                        count = newCallStack.Count;

                        for (int index = 0; index < count; index++)
                        {
                            ICallFrame frame = newCallStack[index];

                            if (frame != null)
                            {
                                if (all && (variableFrame != null) &&
                                    (index == 0) && !WriteLine())
                                {
                                    return false;
                                }

                                if (!WriteCallFrame(
                                        interpreter, frame, null,
                                        NoSeparator, false, false,
                                        false, empty, true))
                                {
                                    return false;
                                }

                                if (all && (frame.Link != null) &&
                                    !WriteCallFrame(
                                        interpreter, frame.Link, null,
                                        NoSeparator, true, false,
                                        false, empty, true))
                                {
                                    return false;
                                }
                            }
                        }
                    }
                    else
                    {
                        //
                        // TODO: Report error here?
                        //
                        return false;
                    }

                    if (newLine && (count > 0) && !WriteLine())
                        return false;

                    return true;
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteCallStackInfo(
            Interpreter interpreter,
            CallStack callStack,
            bool all,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteCallStackInfo(interpreter, callStack, CallStackLimit, all, empty, newLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteCallStackInfo(
            Interpreter interpreter,
            CallStack callStack,
            int limit,
            bool all,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteCallStackInfo(interpreter, callStack, limit, all, empty, newLine,
                CallStackInfoForegroundColor, CallStackInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteCallStackInfo(
            Interpreter interpreter,
            CallStack callStack,
            int limit,
            bool all,
            bool empty, /* NOT USED */
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            StringPairList list = new StringPairList();

            list.Add("CallStack");
            list.Add((IPair<string>)null);

            if ((interpreter != null) && (callStack != null))
            {
                CallStack newCallStack = null;
                Result error = null;

                if (CallFrameOps.Traverse(interpreter, callStack,
                        null, limit, all, ref newCallStack,
                        ref error) == ReturnCode.Ok)
                {
                    int count = newCallStack.Count;

                    for (int index = 0; index < count; index++)
                    {
                        ICallFrame frame = newCallStack[index];

                        if (frame != null)
                        {
                            list.Add(frame.ToString(true));

                            if (all && (frame.Link != null))
                                list.Add(frame.Link.ToString(true));
                        }
                    }
                }
                else
                {
                    list.Add("Error", error);
                }
            }
            else
            {
                list.Add(FormatOps.DisplayNull);
            }

            if (OutputStyle == OutputStyle.Boxed)
                return WriteBox(CallStackInfoBoxName, list, null, false, true,
                    ref hostLeft, ref hostTop, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.Formatted)
                return Write(list, newLine, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.None)
                return true;
            else
                return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if DEBUGGER
        public virtual bool WriteDebuggerInfo(
            Interpreter interpreter,
            bool breakpoints,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteDebuggerInfo(interpreter, breakpoints, empty, newLine,
                DebuggerInfoForegroundColor, DebuggerInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteDebuggerInfo(
            Interpreter interpreter,
            bool breakpoints, /* NOT USED */
            bool empty,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            StringPairList list = new StringPairList();

            list.Add("Debugger");
            list.Add((IPair<string>)null);

            if (interpreter != null)
            {
                lock (interpreter.SyncRoot)
                {
                    IDebugger debugger = interpreter.Debugger;

                    if (debugger != null)
                    {
                        list.Add("Debugger", FormatOps.DisplayPresent);

                        if (empty || (debugger.SuspendCount > 0))
                            list.Add("SuspendCount", debugger.SuspendCount.ToString());

                        if (empty || debugger.Enabled)
                            list.Add("Enabled", debugger.Enabled.ToString());

                        if (empty || (debugger.Active > 0))
                            list.Add("Active", debugger.Active.ToString());

                        if (empty || debugger.SingleStep)
                            list.Add("SingleStep", debugger.SingleStep.ToString());

#if BREAKPOINTS
                        if (empty || debugger.BreakOnToken)
                            list.Add("BreakOnToken", debugger.BreakOnToken.ToString());
#endif

                        if (empty || debugger.BreakOnExecute)
                            list.Add("BreakOnExecute", debugger.BreakOnExecute.ToString());

                        if (empty || debugger.BreakOnCancel)
                            list.Add("BreakOnCancel", debugger.BreakOnCancel.ToString());

                        if (empty || debugger.BreakOnError)
                            list.Add("BreakOnError", debugger.BreakOnError.ToString());

                        if (empty || debugger.BreakOnReturn)
                            list.Add("BreakOnReturn", debugger.BreakOnReturn.ToString());

                        if (empty || debugger.BreakOnTest)
                            list.Add("BreakOnTest", debugger.BreakOnTest.ToString());

                        if (empty || debugger.BreakOnExit)
                            list.Add("BreakOnExit", debugger.BreakOnExit.ToString());

                        if (empty || (debugger.Steps > 0))
                            list.Add("Steps", debugger.Steps.ToString());

                        if (empty || (debugger.Types != BreakpointType.None))
                            list.Add("Types", debugger.Types.ToString());

#if BREAKPOINTS
                        if (empty || ((debugger.Breakpoints != null) && (debugger.Breakpoints.Count > 0)))
                            list.Add("Breakpoints", (debugger.Breakpoints != null) ?
                                debugger.Breakpoints.Count.ToString() : FormatOps.DisplayNull);
#endif

                        if (empty || !String.IsNullOrEmpty(debugger.Command))
                            list.Add("Command", FormatOps.DisplayString(
                                FormatOps.ReplaceNewLines(FormatOps.NormalizeNewLines(
                                    debugger.Command))));

                        if (empty || !String.IsNullOrEmpty(debugger.Result))
                            list.Add("Result", FormatOps.DisplayString(
                                FormatOps.ReplaceNewLines(FormatOps.NormalizeNewLines(
                                    debugger.Result))));

                        Interpreter debugInterpreter = debugger.Interpreter;

                        if (empty || (debugInterpreter != null))
                            list.Add("Interpreter", (debugInterpreter != null) ?
                                debugInterpreter.ToString() : FormatOps.DisplayNull);

                        IHost debugHost = (debugInterpreter != null) ?
                            debugInterpreter.Host : null;

                        if (empty || (debugHost != null))
                            list.Add("Host", (debugHost != null) ?
                                debugHost.ToString() : FormatOps.DisplayNull);

                        if (empty || (debugInterpreter != null))
                            list.Add("NativeThread", (debugInterpreter != null) ?
                                debugInterpreter.ThreadId.ToString() : FormatOps.DisplayNull);

                        if (empty || ((debugInterpreter != null) && (debugInterpreter.Thread != null)))
                            list.Add("ManagedThread", (debugInterpreter != null) &&
                                (debugInterpreter.Thread != null) ?
                                    debugInterpreter.Thread.ManagedThreadId.ToString() :
                                    FormatOps.DisplayNull);

                        if (empty || ((debugInterpreter != null) && (debugInterpreter.VariableEvent != null)))
                            list.Add("VariableEvent", (debugInterpreter != null) &&
                                (debugInterpreter.VariableEvent != null) ?
                                    debugInterpreter.VariableEvent.Handle.ToString() :
                                    FormatOps.DisplayNull);

#if HISTORY
                        if (empty || ((debugInterpreter != null) && debugInterpreter.History))
                            list.Add("History", FormatOps.EnabledAndValue(
                                (debugInterpreter != null) && debugInterpreter.History,
                                (debugInterpreter != null) && (debugInterpreter._History != null) ?
                                    debugInterpreter._History.Count.ToString() :
                                    FormatOps.DisplayNull));

                        if (empty || ((debugInterpreter != null) && (debugInterpreter.HistoryLoadData != null)))
                            list.Add("HistoryLoadData", (debugInterpreter != null) &&
                                (debugInterpreter.HistoryLoadData != null) ?
                                debugInterpreter.HistoryLoadData.ToString() : FormatOps.DisplayNull);

                        if (empty || ((debugInterpreter != null) && (debugInterpreter.HistorySaveData != null)))
                            list.Add("HistorySaveData", (debugInterpreter != null) &&
                                (debugInterpreter.HistorySaveData != null) ?
                                debugInterpreter.HistorySaveData.ToString() : FormatOps.DisplayNull);

                        if (empty || ((debugInterpreter != null) && (debugInterpreter.HistoryEngineFilter != null)))
                            list.Add("HistoryEngineFilter", (debugInterpreter != null) &&
                                (debugInterpreter.HistoryEngineFilter != null) ?
                                debugInterpreter.HistoryEngineFilter.ToString() : FormatOps.DisplayNull);

                        if (empty || ((debugInterpreter != null) && (debugInterpreter.HistoryInfoFilter != null)))
                            list.Add("HistoryInfoFilter", (debugInterpreter != null) &&
                                (debugInterpreter.HistoryInfoFilter != null) ?
                                debugInterpreter.HistoryInfoFilter.ToString() : FormatOps.DisplayNull);

                        if (empty || ((debugInterpreter != null) && (debugInterpreter.HistoryLoadFilter != null)))
                            list.Add("HistoryLoadFilter", (debugInterpreter != null) &&
                                (debugInterpreter.HistoryLoadFilter != null) ?
                                debugInterpreter.HistoryLoadFilter.ToString() : FormatOps.DisplayNull);

                        if (empty || ((debugInterpreter != null) && (debugInterpreter.HistorySaveFilter != null)))
                            list.Add("HistorySaveFilter", (debugInterpreter != null) &&
                                (debugInterpreter.HistorySaveFilter != null) ?
                                debugInterpreter.HistorySaveFilter.ToString() : FormatOps.DisplayNull);
#endif
                    }
                    else
                    {
                        list.Add("Debugger", FormatOps.DisplayNull);
                    }
                }
            }
            else
            {
                list.Add("Debugger", FormatOps.DisplayNull);
            }

            if (OutputStyle == OutputStyle.Boxed)
                return WriteBox(DebuggerInfoBoxName, list, null, false, true,
                    ref hostLeft, ref hostTop, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.Formatted)
                return Write(list, newLine, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.None)
                return true;
            else
                return false;
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteFlagInfo(
            Interpreter interpreter,
            EngineFlags engineFlags,
            SubstitutionFlags substitutionFlags,
            EventFlags eventFlags,
            ExpressionFlags expressionFlags,
            HeaderFlags headerFlags,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteFlagInfo(
                interpreter, engineFlags, substitutionFlags, eventFlags,
                expressionFlags, headerFlags, empty, newLine,
                FlagInfoForegroundColor, FlagInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteFlagInfo(
            Interpreter interpreter,
            EngineFlags engineFlags,
            SubstitutionFlags substitutionFlags,
            EventFlags eventFlags,
            ExpressionFlags expressionFlags,
            HeaderFlags headerFlags,
            bool empty,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            StringPairList list = new StringPairList();

            list.Add("Flags");
            list.Add((IPair<string>)null);

            int headerCount = list.Count;

            if (empty || (engineFlags != EngineFlags.None))
                list.Add("EngineFlags", engineFlags.ToString());

            if (empty || (substitutionFlags != SubstitutionFlags.None))
                list.Add("SubstitutionFlags", substitutionFlags.ToString());

            if (empty || (eventFlags != EventFlags.None))
                list.Add("EventFlags", eventFlags.ToString());

            if (empty || (expressionFlags != ExpressionFlags.None))
                list.Add("ExpressionFlags", expressionFlags.ToString());

            if (empty || (headerFlags != HeaderFlags.None))
                list.Add("HeaderFlags", headerFlags.ToString());

            if (interpreter != null)
            {
                lock (interpreter.SyncRoot)
                {
                    if (empty || ((interpreter != null) && (interpreter.CreateFlags != CreateFlags.None)))
                        list.Add("iCreateFlags", (interpreter != null) ?
                            interpreter.CreateFlags.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.ScriptFlags != ScriptFlags.None)))
                        list.Add("iScriptFlags", (interpreter != null) ?
                            interpreter.ScriptFlags.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.InterpreterFlags != InterpreterFlags.None)))
                        list.Add("iInterpreterFlags", (interpreter != null) ?
                            interpreter.InterpreterFlags.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.EngineFlags != EngineFlags.None)))
                        list.Add("iEngineFlags", (interpreter != null) ?
                            interpreter.EngineFlags.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.SubstitutionFlags != SubstitutionFlags.None)))
                        list.Add("iSubstitutionFlags", (interpreter != null) ?
                            interpreter.SubstitutionFlags.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.ExpressionFlags != ExpressionFlags.None)))
                        list.Add("iExpressionFlags", (interpreter != null) ?
                            interpreter.ExpressionFlags.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.HeaderFlags != HeaderFlags.None)))
                        list.Add("iHeaderFlags", (interpreter != null) ?
                            interpreter.HeaderFlags.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.AfterEventFlags != EventFlags.None)))
                        list.Add("iAfterEventFlags", (interpreter != null) ?
                            interpreter.AfterEventFlags.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.EngineEventFlags != EventFlags.None)))
                        list.Add("iEngineEventFlags", (interpreter != null) ?
                            interpreter.EngineEventFlags.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.QueueEventFlags != EventFlags.None)))
                        list.Add("iQueueEventFlags", (interpreter != null) ?
                            interpreter.QueueEventFlags.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.ServiceEventFlags != EventFlags.None)))
                        list.Add("iServiceEventFlags", (interpreter != null) ?
                            interpreter.ServiceEventFlags.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.UpdateEventFlags != EventFlags.None)))
                        list.Add("iUpdateEventFlags", (interpreter != null) ?
                            interpreter.UpdateEventFlags.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.WaitEventFlags != EventFlags.None)))
                        list.Add("iWaitEventFlags", (interpreter != null) ?
                            interpreter.WaitEventFlags.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.PluginFlags != PluginFlags.None)))
                        list.Add("iPluginFlags", (interpreter != null) ?
                            interpreter.PluginFlags.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.InteractiveEngineFlags != EngineFlags.None)))
                        list.Add("idEngineFlags", (interpreter != null) ?
                            interpreter.InteractiveEngineFlags.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.InteractiveSubstitutionFlags != SubstitutionFlags.None)))
                        list.Add("idSubstitutionFlags", (interpreter != null) ?
                            interpreter.InteractiveSubstitutionFlags.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.InteractiveEventFlags != EventFlags.None)))
                        list.Add("idEventFlags", (interpreter != null) ?
                            interpreter.InteractiveEventFlags.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.InteractiveExpressionFlags != ExpressionFlags.None)))
                        list.Add("idExpressionFlags", (interpreter != null) ?
                            interpreter.InteractiveExpressionFlags.ToString() : FormatOps.DisplayNull);

#if NOTIFY || NOTIFY_OBJECT
                    if (empty || ((interpreter != null) && (interpreter.NotifyTypes != NotifyType.None)))
                        list.Add("iNotifyTypes", (interpreter != null) ?
                            interpreter.NotifyTypes.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter != null) && (interpreter.NotifyFlags != NotifyFlags.None)))
                        list.Add("iNotifyFlags", (interpreter != null) ?
                            interpreter.NotifyFlags.ToString() : FormatOps.DisplayNull);
#endif
                }
            }

            if (list.Count == headerCount)
                list.Add(FormatOps.DisplayEmpty);

            if (OutputStyle == OutputStyle.Boxed)
                return WriteBox(FlagInfoBoxName, list, null, false, true,
                    ref hostLeft, ref hostTop, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.Formatted)
                return Write(list, newLine, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.None)
                return true;
            else
                return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteHostInfo(
            Interpreter interpreter,
            bool dimensions,
            bool formatting,
            bool colors,
            bool names,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteHostInfo(interpreter, dimensions, formatting, colors, names,
                empty, newLine, HostInfoForegroundColor, HostInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteHostInfo(
            Interpreter interpreter,
            bool dimensions,
            bool formatting,
            bool colors,
            bool names,  /* NOT USED */
            bool empty,  /* NOT USED */
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            StringPairList list = new StringPairList();

            list.Add("Host");
            list.Add((IPair<string>)null);

            if (interpreter != null)
            {
#if ISOLATED_PLUGINS
                IHost coreHost = interpreter.CoreHost;

                if (coreHost != null)
                {
                    list.Add("CoreHost", GetHostType(coreHost));
                }
                else
                {
                    list.Add("CoreHost", FormatOps.DisplayNull);
                }
#endif

                IHost host = interpreter.Host;

                if (host != null)
                {
                    list.Add("Host", GetHostType(host));
                    list.Add("Id", GetHostInfo(host, "Id"));
                    list.Add("Kind", GetHostInfo(host,"Kind"));
                    list.Add("Name", GetHostInfo(host, "Name"));
                    list.Add("Description", GetHostInfo(host, "Description"));

                    list.Add("CanExit", GetHostInfo(host, "CanExit"));
                    list.Add("CanForceExit", GetHostInfo(host, "CanForceExit"));
                    list.Add("Exiting", GetHostInfo(host, "Exiting"));

                    list.Add("Profile", GetHostInfo(host, "Profile"));
                    list.Add("NoColor", GetHostInfo(host, "NoColor"));
                    list.Add("NoTitle", GetHostInfo(host, "NoTitle"));
                    list.Add("NoIcon", GetHostInfo(host, "NoIcon"));
                    list.Add("NoProfile", GetHostInfo(host, "NoProfile"));
                    list.Add("NoCancel", GetHostInfo(host, "NoCancel"));
                    list.Add("Echo", GetHostInfo(host, "Echo"));

                    list.Add("IsOpen", GetHostInfo(host, "IsOpen"));
                    list.Add("IsIdle", GetHostInfo(host, "IsIdle"));

                    list.Add("InputEncoding", GetHostInfo(host, "InputEncoding"));
                    list.Add("OutputEncoding", GetHostInfo(host, "OutputEncoding"));
                    list.Add("ErrorEncoding", GetHostInfo(host, "ErrorEncoding"));

                    list.Add("DefaultTitle", GetHostInfo(host, "DefaultTitle"));
                    list.Add("Title", GetHostInfo(host, "Title"));

                    list.Add("Flags", GetHostInfo(host, "GetFlags"));
                    list.Add("TestFlags", GetHostInfo(host, "GetTestFlags"));

                    list.Add("IsInputRedirected", GetHostInfo(host, "IsInputRedirected"));
                    list.Add("IsOutputRedirected", GetHostInfo(host, "IsOutputRedirected"));
                    list.Add("IsErrorRedirected", GetHostInfo(host, "IsErrorRedirected"));

                    _Hosts.Default defaultHost = host as _Hosts.Default;

                    if (defaultHost != null)
                    {
                        if (dimensions)
                        {
                            list.Add((IPair<string>)null);
                            list.Add("Dimensions");
                            list.Add((IPair<string>)null);

                            list.Add("HostLeft", defaultHost.HostLeft.ToString());
                            list.Add("HostTop", defaultHost.HostTop.ToString());
                            list.Add("WindowWidth", defaultHost.WindowWidth.ToString());
                            list.Add("WindowHeight", defaultHost.WindowHeight.ToString());
                            list.Add("ContentWidth", defaultHost.ContentWidth.ToString());
                            list.Add("ContentMargin", defaultHost.ContentMargin.ToString());
                            list.Add("ContentThreshold", defaultHost.ContentThreshold.ToString());
                            list.Add("BoxWidth", defaultHost.BoxWidth.ToString());
                            list.Add("BoxMargin", defaultHost.BoxMargin.ToString());
                        }

                        if (formatting)
                        {
                            list.Add((IPair<string>)null);
                            list.Add("Formatting");
                            list.Add((IPair<string>)null);

                            list.Add("SectionsPerRow", defaultHost.SectionsPerRow.ToString());
                            list.Add("CallStackLimit", defaultHost.CallStackLimit.ToString());
                            list.Add("HistoryLimit", defaultHost.HistoryLimit.ToString());

                            list.Add("Debug", defaultHost.Debug.ToString());
                            list.Add("Exceptions", defaultHost.Exceptions.ToString());
                            list.Add("ReplaceNewLines", defaultHost.ReplaceNewLines.ToString());
                            list.Add("Ellipsis", defaultHost.Ellipsis.ToString());

                            list.Add("NameValueFormat", (defaultHost.NameValueFormat != null) ?
                                defaultHost.NameValueFormat : FormatOps.DisplayNull);

                            list.Add("GoPrompt", (defaultHost.GoPrompt != null) ?
                                defaultHost.GoPrompt : FormatOps.DisplayNull);
                            list.Add("StopPrompt", (defaultHost.StopPrompt != null) ?
                                defaultHost.StopPrompt : FormatOps.DisplayNull);

                            list.Add("BoxCharacterSet",
                                defaultHost.BoxCharacterSet.ToString());

                            list.Add("BoxCharacterSets",
                                defaultHost.BoxCharacterSets.ToRawString());
                        }

                        //
                        // NOTE: Check if they want to output the color settings.
                        //
                        if (colors)
                        {
                            list.Add((IPair<string>)null);
                            list.Add("Colors");
                            list.Add((IPair<string>)null);

                            List<PropertyInfo> propertyList = GetColorProperties(
                                GetType(), MatchMode.None, null, false, true, false);

                            foreach (PropertyInfo propertyInfo in propertyList)
                            {
                                if (propertyInfo != null)
                                {
                                    object value = propertyInfo.GetValue(this, null);

                                    list.Add(propertyInfo.Name, (value != null) ?
                                        value.ToString() : FormatOps.DisplayNull);
                                }
                            }
                        }
                    }
                }
                else
                {
                    list.Add("Host", FormatOps.DisplayNull);
                }
            }
            else
            {
                list.Add("Host", FormatOps.DisplayNull);
            }

            if (OutputStyle == OutputStyle.Boxed)
                return WriteBox(HostInfoBoxName, list, null, false, true,
                    ref hostLeft, ref hostTop, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.Formatted)
                return Write(list, newLine, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.None)
                return true;
            else
                return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteInterpreterInfo(
            Interpreter interpreter,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteInterpreterInfo(interpreter, empty, newLine,
                InterpreterInfoForegroundColor, InterpreterInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteInterpreterInfo(
            Interpreter interpreter,
            bool empty,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            StringPairList list = new StringPairList();

            list.Add("Interpreter");
            list.Add((IPair<string>)null);

            int headerCount = list.Count;

            if (interpreter != null)
            {
                lock (interpreter.SyncRoot)
                {
                    list.Add("Id", interpreter.ToString()); // NOTE: Also in engine info.

                    if (empty || interpreter.Disposed)
                        list.Add("Disposed", interpreter.Disposed.ToString());

                    if (empty || interpreter.ThrowOnFeatureNotSupported)
                        list.Add("ThrowOnFeatureNotSupported",
                            interpreter.ThrowOnFeatureNotSupported.ToString());

                    if (empty || (interpreter.TraceTextWriter != null))
                        list.Add("TraceTextWriter", (interpreter.TraceTextWriter != null) ?
                            interpreter.TraceTextWriter.ToString() : FormatOps.DisplayNull);

                    if (empty || (interpreter.DebugTextWriter != null))
                        list.Add("DebugTextWriter", (interpreter.DebugTextWriter != null) ?
                            interpreter.DebugTextWriter.ToString() : FormatOps.DisplayNull);

                    if (empty || (interpreter.ApplicationObject != null))
                        list.Add("ApplicationObject", (interpreter.ApplicationObject != null) ?
                            interpreter.ApplicationObject.ToString() : FormatOps.DisplayNull);

                    if (empty || (interpreter.PolicyObject != null))
                        list.Add("PolicyObject", (interpreter.PolicyObject != null) ?
                            interpreter.PolicyObject.ToString() : FormatOps.DisplayNull);

                    if (empty || (interpreter.ResolverObject != null))
                        list.Add("ResolverObject", (interpreter.ResolverObject != null) ?
                            interpreter.ResolverObject.ToString() : FormatOps.DisplayNull);

                    if (empty || (interpreter.UserObject != null))
                        list.Add("UserObject", (interpreter.UserObject != null) ?
                            interpreter.UserObject.ToString() : FormatOps.DisplayNull);

                    bool sameAppDomain = RuntimeOps.IsSameAppDomain(interpreter);

#if DEBUGGER
                    if (sameAppDomain) // NOTE: Origin of delegate unknown.
                    {
                        InteractiveLoopCallback interactiveLoopCallback = interpreter.InteractiveLoopCallback;

                        if (empty || ((interactiveLoopCallback != null) && (interactiveLoopCallback.Method != null)))
                            list.Add("InteractiveLoopCallback",
                                ((interactiveLoopCallback != null) && (interactiveLoopCallback.Method != null)) ?
                                FormatOps.DelegateMethodName(interactiveLoopCallback.Method.DeclaringType,
                                    interactiveLoopCallback.Method.Name, false) : FormatOps.DisplayNull);
                    }
#endif

                    if (sameAppDomain) // NOTE: Non-public members.
                    {
                        if (empty || (interpreter.masterInterpreter != null))
                            list.Add("MasterInterpreter", (interpreter.masterInterpreter != null) ?
                                interpreter.masterInterpreter.ToString() : FormatOps.DisplayNull);

                        if (empty || !String.IsNullOrEmpty(interpreter.slaveName))
                            list.Add("SlaveName",
                                FormatOps.DisplayString(interpreter.slaveName));
                    }

                    if (empty || !String.IsNullOrEmpty(interpreter.LibraryPath))
                        list.Add("LibraryPath", FormatOps.DisplayString(interpreter.LibraryPath));

                    if (empty || ((interpreter.AutoPathList != null) && (interpreter.AutoPathList.Count > 0)))
                        list.Add("AutoPathList", FormatOps.DisplayList(interpreter.AutoPathList));

                    if (empty || !String.IsNullOrEmpty(interpreter.Unknown))
                        list.Add("Unknown",
                            FormatOps.DisplayString(interpreter.Unknown));

                    if (empty || !String.IsNullOrEmpty(interpreter.PackageUnknown))
                        list.Add("PackageUnknown",
                            FormatOps.DisplayString(interpreter.PackageUnknown));

                    if (empty || !String.IsNullOrEmpty(interpreter.BackgroundError))
                        list.Add("BackgroundError",
                            FormatOps.DisplayString(interpreter.BackgroundError));

                    if (sameAppDomain) // NOTE: Non-public member.
                    {
                        if (empty || !String.IsNullOrEmpty(interpreter.initializedPath))
                            list.Add("InitializedPath",
                                FormatOps.DisplayString(interpreter.initializedPath));
                    }

                    if (empty || !String.IsNullOrEmpty(interpreter.TestPath))
                        list.Add("TestPath",
                            FormatOps.DisplayString(interpreter.TestPath));

                    if (sameAppDomain) // NOTE: Non-serializable member.
                    {
                        if (empty || (interpreter.Binder != null))
                            list.Add("Binder", (interpreter.Binder != null) ?
                                interpreter.Binder.ToString() : FormatOps.DisplayNull);
                    }

                    if (empty || (interpreter.CultureInfo != null))
                        list.Add("CultureInfo", (interpreter.CultureInfo != null) ?
                            FormatOps.CultureName(interpreter.CultureInfo, true) :
                            FormatOps.DisplayNull);

                    if (empty || !String.IsNullOrEmpty(interpreter.InteractiveInput))
                        list.Add("InteractiveInput",
                            FormatOps.DisplayString(FormatOps.ReplaceNewLines(
                                FormatOps.NormalizeNewLines(interpreter.InteractiveInput))));

                    if (empty || !String.IsNullOrEmpty(interpreter.PreviousInteractiveInput))
                        list.Add("PreviousInteractiveInput",
                            FormatOps.DisplayString(FormatOps.ReplaceNewLines(
                                FormatOps.NormalizeNewLines(interpreter.PreviousInteractiveInput))));

                    if (empty || (interpreter.ScriptLocation != null))
                        list.Add("ScriptLocation (Override)",
                            (interpreter.ScriptLocation != null) ?
                            interpreter.ScriptLocation.ToString() :
                            FormatOps.DisplayNull);

                    if (empty ||
                        (!interpreter.ScriptLocations.IsEmpty &&
                         (interpreter.ScriptLocations.Peek() != null)))
                    {
                        list.Add("ScriptLocation (Actual)",
                            !interpreter.ScriptLocations.IsEmpty ?
                            interpreter.ScriptLocations.Peek().ToString() :
                            FormatOps.DisplayNull);
                    }

                    if (empty || (interpreter.PreviousProcessId != 0))
                        list.Add("PreviousProcessId", interpreter.PreviousProcessId.ToString());

#if CACHE_STATISTICS
#if EXECUTE_CACHE
                    if (sameAppDomain) // NOTE: Non-public member.
                    {
                        if (empty || ((interpreter.executeCache != null) && (interpreter.executeCache.Count > 0)))
                            list.Add("ExecuteCache", (interpreter.executeCache != null) ?
                                interpreter.executeCache.CacheToString() : FormatOps.DisplayNull);
                    }
#endif

#if LIST_CACHE
                    if (sameAppDomain) // NOTE: Non-public member.
                    {
                        if (empty || ((interpreter.stringListCache != null) && (interpreter.stringListCache.Count > 0)))
                            list.Add("StringListCache", (interpreter.stringListCache != null) ?
                                interpreter.stringListCache.CacheToString() : FormatOps.DisplayNull);
                    }
#endif

#if PARSE_CACHE
                    if (sameAppDomain) // NOTE: Non-public member.
                    {
                        if (empty || ((interpreter.parseStateCache != null) && (interpreter.parseStateCache.Count > 0)))
                            list.Add("ParseStateCache", (interpreter.parseStateCache != null) ?
                                interpreter.parseStateCache.CacheToString() : FormatOps.DisplayNull);
                    }
#endif

                    #region Dead Code
#if DEAD_CODE
                    if (sameAppDomain) // NOTE: Non-public member.
                    {
                        if (empty || ((interpreter.expressionStateCache != null) && (interpreter.expressionStateCache.Count > 0)))
                            list.Add("ExpressionStateCache", (interpreter.expressionStateCache != null) ?
                                interpreter.expressionStateCache.CacheToString() : FormatOps.DisplayNull);
                    }
#endif
                    #endregion

#if TYPE_CACHE
                    if (sameAppDomain) // NOTE: Non-public member.
                    {
                        if (empty || ((interpreter.typeCache != null) && (interpreter.typeCache.Count > 0)))
                            list.Add("TypeCache", (interpreter.typeCache != null) ?
                                interpreter.typeCache.CacheToString() : FormatOps.DisplayNull);
                    }
#endif

#if COM_TYPE_CACHE
                    if (sameAppDomain) // NOTE: Non-public member.
                    {
                        if (empty || ((interpreter.comTypeListCache != null) && (interpreter.comTypeListCache.Count > 0)))
                            list.Add("ComTypeListCache", (interpreter.comTypeListCache != null) ?
                                interpreter.comTypeListCache.CacheToString() : FormatOps.DisplayNull);
                    }
#endif
#endif
                }

                if (list.Count == headerCount)
                    list.Add(FormatOps.DisplayEmpty);
            }
            else
            {
                list.Add(FormatOps.DisplayNull);
            }

            if (OutputStyle == OutputStyle.Boxed)
                return WriteBox(InterpreterInfoBoxName, list, null, false, true,
                    ref hostLeft, ref hostTop, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.Formatted)
                return Write(list, newLine, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.None)
                return true;
            else
                return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteEngineInfo(
            Interpreter interpreter,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteEngineInfo(interpreter, empty, newLine,
                EngineInfoForegroundColor, EngineInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteEngineInfo(
            Interpreter interpreter,
            bool empty,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            if (interpreter != null)
            {
                StringPairList list = new StringPairList();

                list.Add("Engine");
                list.Add((IPair<string>)null);

                list.Add("ThrowOnDisposed", Engine.IsThrowOnDisposed(null, false).ToString());

#if NATIVE && WINDOWS
                list.Add("StackExtraSpace", Engine.GetStackExtraSpace().ToString());
#endif

                list.Add("NativeThread", GlobalState.GetCurrentThreadId().ToString());
                list.Add("ManagedThread", Thread.CurrentThread.ManagedThreadId.ToString());
                list.Add("PrimaryNativeThread", GlobalState.GetPrimaryThreadId().ToString());

                Thread primaryThread = GlobalState.GetPrimaryThread();

                list.Add("PrimaryManagedThread", (primaryThread != null) ?
                    primaryThread.ManagedThreadId.ToString() : FormatOps.DisplayNull);

                NewHostCallback newHostCallback = Interpreter.NewHostCallback;

                if (empty || ((newHostCallback != null) && (newHostCallback.Method != null)))
                    list.Add("NewHostCallback",
                        ((newHostCallback != null) && (newHostCallback.Method != null)) ?
                        FormatOps.DelegateMethodName(newHostCallback.Method.DeclaringType,
                            newHostCallback.Method.Name, false) : FormatOps.DisplayNull);

                lock (interpreter.SyncRoot)
                {
                    list.Add("InterpreterId", interpreter.ToString());

                    if (empty || (interpreter.ThreadStackSize != 0))
                        list.Add("ThreadStackSize", interpreter.ThreadStackSize.ToString());

                    if (empty || (interpreter.InteractiveLoops > 0))
                        list.Add("InteractiveLoops", interpreter.InteractiveLoops.ToString());

                    if (empty || (interpreter.CommandCount > 0))
                        list.Add("CommandCount", interpreter.CommandCount.ToString());

                    if (empty || (interpreter.CommandDecision != PolicyDecision.None))
                        list.Add("CommandDecision", interpreter.CommandDecision.ToString());

                    if (empty || (interpreter.ScriptDecision != PolicyDecision.None))
                        list.Add("ScriptDecision", interpreter.ScriptDecision.ToString());

                    if (empty || (interpreter.FileDecision != PolicyDecision.None))
                        list.Add("FileDecision", interpreter.FileDecision.ToString());

                    if (empty || (interpreter.StreamDecision != PolicyDecision.None))
                        list.Add("StreamDecision", interpreter.StreamDecision.ToString());

#if POLICY_TRACE
                    if (empty || interpreter.PolicyTrace)
                        list.Add("PolicyTrace", interpreter.PolicyTrace.ToString());
#endif

                    if (empty || ((interpreter.TrustedPaths != null) && (interpreter.TrustedPaths.Count > 0)))
                        list.Add("TrustedPaths", FormatOps.DisplayList(interpreter.TrustedPaths));

                    if (empty || ((interpreter.TrustedTypes != null) && (interpreter.TrustedTypes.Count > 0)))
                        list.Add("TrustedTypes", FormatOps.DisplayKeys(interpreter.TrustedTypes));

                    list.Add("ReturnCode", interpreter.ReturnCode.ToString());

                    if (empty || (interpreter.ErrorLine != 0))
                        list.Add("ErrorLine", interpreter.ErrorLine.ToString());

                    if (empty || (interpreter.ErrorCode != null))
                        list.Add("ErrorCode", (interpreter.ErrorCode != null) ?
                            interpreter.ErrorCode : FormatOps.DisplayNull);

                    //if (empty || (interpreter.ErrorInfo != null))
                    //    list.Add("ErrorInfo", (interpreter.ErrorInfo != null) ?
                    //        interpreter.ErrorInfo : FormatOps.DisplayNull);

                    if (empty || (interpreter.ErrorFrames != 0))
                        list.Add("ErrorFrames", interpreter.ErrorFrames.ToString());

                    if (empty || interpreter.StackOverflow)
                        list.Add("StackOverflow", interpreter.StackOverflow.ToString());

                    if (empty || (interpreter.Levels > 0))
                        list.Add("Levels", interpreter.Levels.ToString());

                    if (empty || (interpreter.PreviousLevels > 0))
                        list.Add("PreviousLevels", interpreter.PreviousLevels.ToString());

                    if (empty || (interpreter.ParserLevels > 0))
                        list.Add("ParserLevels", interpreter.ParserLevels.ToString());

                    if (empty || (interpreter.ExpressionLevels > 0))
                        list.Add("ExpressionLevels", interpreter.ExpressionLevels.ToString());

                    if (empty || (interpreter.UnknownLevels > 0))
                        list.Add("UnknownLevels", interpreter.UnknownLevels.ToString());

                    if (empty || (interpreter.PolicyLevels > 0))
                        list.Add("PolicyLevels", interpreter.PolicyLevels.ToString());

                    if (empty || (interpreter.TraceLevels > 0))
                        list.Add("TraceLevels", interpreter.TraceLevels.ToString());

#if NOTIFY || NOTIFY_OBJECT
                    if (empty || (interpreter.NotifyLevels > 0))
                        list.Add("NotifyLevels", interpreter.NotifyLevels.ToString());
#endif

                    if (empty || (interpreter.Precision != TclVars.PrecisionValue))
                        list.Add("Precision", interpreter.Precision.ToString());

#if XML
                    if (empty || interpreter.AllXml)
                        list.Add("AllXml", interpreter.AllXml.ToString());

                    if (empty || interpreter.ValidateXml)
                        list.Add("ValidateXml", interpreter.ValidateXml.ToString());
#endif
                }

#if NATIVE && TCL && NATIVE_PACKAGE
                NativePackage.AddInfo(empty, ref list);
#endif

                if (OutputStyle == OutputStyle.Boxed)
                    return WriteBox(EngineInfoBoxName, list, null, false, true,
                        ref hostLeft, ref hostTop, foregroundColor, backgroundColor);
                else if (OutputStyle == OutputStyle.Formatted)
                    return Write(list, newLine, foregroundColor, backgroundColor);
                else if (OutputStyle == OutputStyle.None)
                    return true;
                else
                    return false;
            }
            else
            {
                return true;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteEntityInfo(
            Interpreter interpreter,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteEntityInfo(interpreter, empty, newLine,
                EntityInfoForegroundColor, EntityInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteEntityInfo(
            Interpreter interpreter,
            bool empty,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            if ((interpreter != null) &&
                RuntimeOps.IsSameAppDomain(interpreter)) // NOTE: Non-public members (mostly).
            {
                StringPairList list = new StringPairList();

                list.Add("Entity");
                list.Add((IPair<string>)null);

                int headerCount = list.Count;

                lock (interpreter.SyncRoot)
                {
                    if (empty || ((interpreter.slaveInterpreters != null) && (interpreter.slaveInterpreters.Count > 0)))
                        list.Add("SlaveInterpreters", (interpreter.slaveInterpreters != null) ?
                            interpreter.slaveInterpreters.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.packages != null) && (interpreter.packages.Count > 0)))
                        list.Add("Packages", (interpreter.packages != null) ?
                            interpreter.packages.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.packageIndexes != null) && (interpreter.packageIndexes.Count > 0)))
                        list.Add("PackageIndexes", (interpreter.packageIndexes != null) ?
                            interpreter.packageIndexes.Count.ToString() : FormatOps.DisplayNull);

#if APPDOMAINS
                    if (empty || ((interpreter.appDomains != null) && (interpreter.appDomains.Count > 0)))
                        list.Add("AppDomains", (interpreter.appDomains != null) ?
                            interpreter.appDomains.Count.ToString() : FormatOps.DisplayNull);
#endif

                    if (empty || ((interpreter.plugins != null) && (interpreter.plugins.Count > 0)))
                        list.Add("Plugins", (interpreter.plugins != null) ?
                            interpreter.plugins.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.commands != null) && (interpreter.commands.Count > 0)))
                        list.Add("Commands", (interpreter.commands != null) ?
                            interpreter.commands.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.hiddenCommands != null) && (interpreter.hiddenCommands.Count > 0)))
                        list.Add("HiddenCommands", (interpreter.hiddenCommands != null) ?
                            interpreter.hiddenCommands.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.aliases != null) && (interpreter.aliases.Count > 0)))
                        list.Add("Aliases", (interpreter.aliases != null) ?
                            interpreter.aliases.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.procedures != null) && (interpreter.procedures.Count > 0)))
                        list.Add("Procedures", (interpreter.procedures != null) ?
                            interpreter.procedures.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.hiddenProcedures != null) && (interpreter.hiddenProcedures.Count > 0)))
                        list.Add("HiddenProcedures", (interpreter.hiddenProcedures != null) ?
                            interpreter.hiddenProcedures.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.executes != null) && (interpreter.executes.Count > 0)))
                        list.Add("Executes", (interpreter.executes != null) ?
                            interpreter.executes.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.hiddenExecutes != null) && (interpreter.hiddenExecutes.Count > 0)))
                        list.Add("HiddenExecutes", (interpreter.hiddenExecutes != null) ?
                            interpreter.hiddenExecutes.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.functions != null) && (interpreter.functions.Count > 0)))
                        list.Add("Functions", (interpreter.functions != null) ?
                            interpreter.functions.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.operators != null) && (interpreter.operators.Count > 0)))
                        list.Add("Operators", (interpreter.operators != null) ?
                            interpreter.operators.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.policies != null) && (interpreter.policies.Count > 0)))
                        list.Add("Policies", (interpreter.policies != null) ?
                            interpreter.policies.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.scopes != null) && (interpreter.scopes.Count > 0)))
                        list.Add("Scopes", (interpreter.scopes != null) ?
                            interpreter.scopes.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.traces != null) && (interpreter.traces.Count > 0)))
                        list.Add("Traces", (interpreter.traces != null) ?
                            interpreter.traces.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.channels != null) && (interpreter.channels.Count > 0)))
                        list.Add("Channels", (interpreter.channels != null) ?
                            interpreter.channels.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.Encodings != null) && (interpreter.Encodings.Count > 0)))
                        list.Add("Encodings", (interpreter.Encodings != null) ?
                            interpreter.Encodings.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.objects != null) && (interpreter.objects.Count > 0)))
                        list.Add("Objects", (interpreter.objects != null) ?
                            interpreter.objects.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.ObjectTypes != null) && (interpreter.ObjectTypes.Count > 0)))
                        list.Add("ObjectTypes", (interpreter.ObjectTypes != null) ?
                            interpreter.ObjectTypes.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.ObjectNamespaces != null) && (interpreter.ObjectNamespaces.Count > 0)))
                        list.Add("ObjectNamespaces", (interpreter.ObjectNamespaces != null) ?
                            interpreter.ObjectNamespaces.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.ObjectInterfaces != null) && (interpreter.ObjectInterfaces.Count > 0)))
                        list.Add("ObjectInterfaces", (interpreter.ObjectInterfaces != null) ?
                            interpreter.ObjectInterfaces.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.callbacks != null) && (interpreter.callbacks.Count > 0)))
                        list.Add("Callbacks", (interpreter.callbacks != null) ?
                            interpreter.callbacks.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.resolvers != null) && (interpreter.resolvers.Count > 0)))
                        list.Add("Resolvers", (interpreter.resolvers != null) ?
                            interpreter.resolvers.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.ArraySearches != null) && (interpreter.ArraySearches.Count > 0)))
                        list.Add("ArraySearches", (interpreter.ArraySearches != null) ?
                            interpreter.ArraySearches.Count.ToString() : FormatOps.DisplayNull);

#if CALLBACK_QUEUE
                    if (empty || ((interpreter.callbackQueue != null) && (interpreter.callbackQueue.Count > 0)))
                        list.Add("CallbackQueue", (interpreter.callbackQueue != null) ?
                            interpreter.callbackQueue.Count.ToString() : FormatOps.DisplayNull);
#endif

                    IEventManager eventManager = interpreter.EventManager;

                    if (empty || ((eventManager != null) && (eventManager.EventCount > 0)))
                        list.Add("Events", (eventManager != null) ?
                            eventManager.EventCount.ToString() : FormatOps.DisplayNull);

                    if (empty || ((eventManager != null) && (eventManager.IdleEventCount > 0)))
                        list.Add("IdleEvents", (eventManager != null) ?
                            eventManager.IdleEventCount.ToString() : FormatOps.DisplayNull);

                    if (empty || ((eventManager != null) && (eventManager.TotalEventCount > 0)))
                        list.Add("TotalEvents", (eventManager != null) ?
                            eventManager.TotalEventCount.ToString() : FormatOps.DisplayNull);

                    if (empty || ((eventManager != null) && (eventManager.MaximumEventCount > 0)))
                        list.Add("MaximumEvents", (eventManager != null) ?
                            eventManager.MaximumEventCount.ToString() : FormatOps.DisplayNull);

                    if (empty || ((eventManager != null) && (eventManager.MaximumIdleEventCount > 0)))
                        list.Add("MaximumIdleEvents", (eventManager != null) ?
                            eventManager.MaximumIdleEventCount.ToString() : FormatOps.DisplayNull);

#if DATA
                    if (empty || ((interpreter.connections != null) && (interpreter.connections.Count > 0)))
                        list.Add("DbConnections", (interpreter.connections != null) ?
                            interpreter.connections.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.transactions != null) && (interpreter.transactions.Count > 0)))
                        list.Add("DbTransactions", (interpreter.transactions != null) ?
                            interpreter.transactions.Count.ToString() : FormatOps.DisplayNull);
#endif

#if NATIVE && TCL
                    if (empty || ((interpreter.tclInterps != null) && (interpreter.tclInterps.Count > 0)))
                        list.Add("TclInterps", (interpreter.tclInterps != null) ?
                            interpreter.tclInterps.Count.ToString() : FormatOps.DisplayNull);

#if TCL_THREADS
                    if (empty || ((interpreter.tclThreads != null) && (interpreter.tclThreads.Count > 0)))
                        list.Add("TclThreads", (interpreter.tclThreads != null) ?
                            interpreter.tclThreads.Count.ToString() : FormatOps.DisplayNull);
#endif

                    if (empty || ((interpreter.tclBridges != null) && (interpreter.tclBridges.Count > 0)))
                        list.Add("TclBridges", (interpreter.tclBridges != null) ?
                            interpreter.tclBridges.Count.ToString() : FormatOps.DisplayNull);
#endif

#if NATIVE && LIBRARY
                    if (empty || ((interpreter.modules != null) && (interpreter.modules.Count > 0)))
                        list.Add("Modules", (interpreter.modules != null) ?
                            interpreter.modules.Count.ToString() : FormatOps.DisplayNull);

                    if (empty || ((interpreter.delegates != null) && (interpreter.delegates.Count > 0)))
                        list.Add("Delegates", (interpreter.delegates != null) ?
                            interpreter.delegates.Count.ToString() : FormatOps.DisplayNull);
#endif
                }

                if (list.Count == headerCount)
                    list.Add(FormatOps.DisplayEmpty);

                if (OutputStyle == OutputStyle.Boxed)
                    return WriteBox(EntityInfoBoxName, list, null, false, true,
                        ref hostLeft, ref hostTop, foregroundColor, backgroundColor);
                else if (OutputStyle == OutputStyle.Formatted)
                    return Write(list, newLine, foregroundColor, backgroundColor);
                else if (OutputStyle == OutputStyle.None)
                    return true;
                else
                    return false;
            }
            else
            {
                return true;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteStackInfo(
            Interpreter interpreter,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteStackInfo(interpreter, empty, newLine,
                EngineInfoForegroundColor, EngineInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteStackInfo(
            Interpreter interpreter,
            bool empty, /* NOT USED */
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            StringPairList list = new StringPairList();

            list.Add("Stack");
            list.Add((IPair<string>)null);

            //
            // NOTE: *WARNING* The numbers output here are NOT "live" as they are the
            //       values from the last trip through the interpreter stack checking
            //       code (which may be RADICALLY different from the current usage if
            //       we recently encountered a stack overflow).
            //
            UIntPtr used = UIntPtr.Zero;
            UIntPtr allocated = UIntPtr.Zero;
            UIntPtr extra = UIntPtr.Zero;
            UIntPtr margin = UIntPtr.Zero;
            UIntPtr maximum = UIntPtr.Zero;
            UIntPtr reserve = UIntPtr.Zero;
            UIntPtr commit = UIntPtr.Zero;

            Result error = null;

            if (RuntimeOps.GetStackSize(
                    ref used, ref allocated, ref extra, ref margin, ref maximum,
                    ref reserve, ref commit, ref error) == ReturnCode.Ok)
            {
                list.Add("Used", used.ToString());
                list.Add("Allocated", allocated.ToString());
                list.Add("Extra", extra.ToString());
                list.Add("Margin", margin.ToString());
                list.Add("Maximum", maximum.ToString());
                list.Add("Reserve", reserve.ToString());
                list.Add("Commit", commit.ToString());
            }
            else
            {
                list.Add("Error", error);
            }

            if (OutputStyle == OutputStyle.Boxed)
                return WriteBox(StackInfoBoxName, list, null, false, true,
                    ref hostLeft, ref hostTop, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.Formatted)
                return Write(list, newLine, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.None)
                return true;
            else
                return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteControlInfo(
            Interpreter interpreter,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteControlInfo(interpreter, empty, newLine,
                ControlInfoForegroundColor, ControlInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteControlInfo(
            Interpreter interpreter,
            bool empty,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            if (interpreter != null)
            {
                StringPairList list = new StringPairList();

                list.Add("Control");
                list.Add((IPair<string>)null);

                lock (interpreter.SyncRoot)
                {
                    list.Add("NativeThread", interpreter.ThreadId.ToString());

                    bool sameAppDomain = RuntimeOps.IsSameAppDomain(interpreter);

                    if (sameAppDomain) // NOTE: Non-serializable member.
                    {
                        list.Add("ManagedThread", (interpreter.Thread != null) ?
                            interpreter.Thread.ManagedThreadId.ToString() : FormatOps.DisplayNull);
                    }

                    list.Add("VariableEvent", (interpreter.VariableEvent != null) ?
                        interpreter.VariableEvent.Handle.ToString() : FormatOps.DisplayNull);

                    if (empty || interpreter.Interactive)
                        list.Add("Interactive", interpreter.Interactive.ToString());

                    if (empty || interpreter.Debug)
                        list.Add("Debug", interpreter.Debug.ToString());

                    if (empty || interpreter.Quiet)
                        list.Add("Quiet", interpreter.Quiet.ToString());

                    bool quiet = Interpreter.DefaultQuiet;

                    if (empty || quiet)
                        list.Add("DefaultQuiet", quiet.ToString());

                    quiet = DebugOps.GetDefaultQuiet(quiet);

                    if (empty || quiet)
                        list.Add("GetDefaultQuiet", quiet.ToString());

                    quiet = interpreter.ShouldBeQuiet(quiet);

                    if (empty || quiet)
                        list.Add("ShouldBeQuiet", quiet.ToString());

                    if (empty || interpreter.ReadOnly)
                        list.Add("ReadOnly", interpreter.ReadOnly.ToString());

                    if (empty || (interpreter.RecursionLimit > 0))
                        list.Add("RecursionLimit", interpreter.RecursionLimit.ToString());

                    IEventManager eventManager = interpreter.EventManager;

                    if (empty || ((eventManager != null) && eventManager.Enabled))
                        list.Add("EventManager", (eventManager != null) ?
                            eventManager.Enabled.ToString() : FormatOps.DisplayNull);

                    if (empty || (interpreter.WaitCount > 0))
                        list.Add("WaitCount", interpreter.WaitCount.ToString());

                    if (empty || (interpreter.WaitSpinCount > 0))
                        list.Add("WaitSpinCount", interpreter.WaitSpinCount.ToString());

#if NETWORK
                    if (empty || (interpreter.ServerSockets > 0))
                        list.Add("ServerSockets", interpreter.ServerSockets.ToString());
#endif

                    if (empty || (interpreter.Timeout != _Timeout.Infinite))
                        list.Add("Timeout", interpreter.Timeout.ToString());

                    if (empty || (interpreter.SleepTime != 0))
                        list.Add("SleepTime", interpreter.SleepTime.ToString());

                    if (empty || interpreter.Cancel)
                        list.Add("Cancel", interpreter.Cancel.ToString());

                    if (empty || interpreter.Unwind)
                        list.Add("Unwind", interpreter.Unwind.ToString());

                    if (empty || (interpreter.CancelResult != null))
                        list.Add("CancelResult", (interpreter.CancelResult != null) ?
                            interpreter.CancelResult.ToString() : FormatOps.DisplayNull);

                    if (empty || interpreter.Halt)
                        list.Add("Halt", interpreter.Halt.ToString());

                    if (empty || (interpreter.HaltResult != null))
                        list.Add("HaltResult", (interpreter.HaltResult != null) ?
                            interpreter.HaltResult.ToString() : FormatOps.DisplayNull);

                    if (empty || interpreter.Deleted)
                        list.Add("Deleted", interpreter.Deleted.ToString());

                    if (empty || interpreter.Exit)
                        list.Add("Exit", interpreter.Exit.ToString());

                    list.Add("ExitCode", interpreter.ExitCode.ToString());

#if HISTORY
                    if (empty || interpreter.History)
                        list.Add("History", FormatOps.EnabledAndValue(
                            interpreter.History, (interpreter._History != null) ?
                                interpreter._History.Count.ToString() : FormatOps.DisplayNull));

                    if (empty || (interpreter.HistoryLoadData != null))
                        list.Add("HistoryLoadData", (interpreter.HistoryLoadData != null) ?
                            interpreter.HistoryLoadData.ToString() : FormatOps.DisplayNull);

                    if (empty || (interpreter.HistorySaveData != null))
                        list.Add("HistorySaveData", (interpreter.HistorySaveData != null) ?
                            interpreter.HistorySaveData.ToString() : FormatOps.DisplayNull);

                    if (empty || (interpreter.HistoryEngineFilter != null))
                        list.Add("HistoryEngineFilter", (interpreter.HistoryEngineFilter != null) ?
                            interpreter.HistoryEngineFilter.ToString() : FormatOps.DisplayNull);

                    if (empty || (interpreter.HistoryInfoFilter != null))
                        list.Add("HistoryInfoFilter", (interpreter.HistoryInfoFilter != null) ?
                            interpreter.HistoryInfoFilter.ToString() : FormatOps.DisplayNull);

                    if (empty || (interpreter.HistoryLoadFilter != null))
                        list.Add("HistoryLoadFilter", (interpreter.HistoryLoadFilter != null) ?
                            interpreter.HistoryLoadFilter.ToString() : FormatOps.DisplayNull);

                    if (empty || (interpreter.HistorySaveFilter != null))
                        list.Add("HistorySaveFilter", (interpreter.HistorySaveFilter != null) ?
                            interpreter.HistorySaveFilter.ToString() : FormatOps.DisplayNull);
#endif
                }

                if (OutputStyle == OutputStyle.Boxed)
                    return WriteBox(ControlInfoBoxName, list, null, false, true,
                        ref hostLeft, ref hostTop, foregroundColor, backgroundColor);
                else if (OutputStyle == OutputStyle.Formatted)
                    return Write(list, newLine, foregroundColor, backgroundColor);
                else if (OutputStyle == OutputStyle.None)
                    return true;
                else
                    return false;
            }
            else
            {
                return true;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteTestInfo(
            Interpreter interpreter,
            bool empty,
            bool newLine
            )
        {
            return WriteTestInfo(interpreter, empty, newLine,
                TestInfoForegroundColor, TestInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteTestInfo(
            Interpreter interpreter,
            bool empty,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            StringPairList list = new StringPairList();

            list.Add("Test");
            list.Add((IPair<string>)null);

            int headerCount = list.Count;

            if (interpreter != null)
            {
                lock (interpreter.SyncRoot)
                {
                    if ((interpreter.TestStatistics != null) &&
                        (interpreter.TestStatistics.Length >= (int)TestInformationType.SizeOf))
                    {
                        if (empty || (interpreter.TestStatistics[(int)TestInformationType.Total] > 0))
                            list.Add("Total", interpreter.TestStatistics[(int)TestInformationType.Total].ToString());

                        if (empty || (interpreter.TestStatistics[(int)TestInformationType.Skipped] > 0))
                            list.Add("Skipped", interpreter.TestStatistics[(int)TestInformationType.Skipped].ToString());

                        if (empty || (interpreter.TestStatistics[(int)TestInformationType.Passed] > 0))
                            list.Add("Passed", interpreter.TestStatistics[(int)TestInformationType.Passed].ToString());

                        if (empty || (interpreter.TestStatistics[(int)TestInformationType.Failed] > 0))
                            list.Add("Failed", interpreter.TestStatistics[(int)TestInformationType.Failed].ToString());
                    }

                    if (empty || ((interpreter.TestMatch != null) && (interpreter.TestMatch.Count > 0)))
                        list.Add("Match", FormatOps.DisplayList(interpreter.TestMatch));

                    if (empty || ((interpreter.TestSkip != null) && (interpreter.TestSkip.Count > 0)))
                        list.Add("Skip", FormatOps.DisplayList(interpreter.TestSkip));

#if DEBUGGER
                    if (empty || ((interpreter.TestBreakpoints != null) && (interpreter.TestBreakpoints.Count > 0)))
                        list.Add("Breakpoints", (interpreter.TestBreakpoints != null) ?
                            FormatOps.DisplayKeys(interpreter.TestBreakpoints) : FormatOps.DisplayNull);
#endif

                    if (empty || ((interpreter.TestConstraints != null) && (interpreter.TestConstraints.Count > 0)))
                        list.Add("Constraints", (interpreter.TestConstraints != null) ?
                            FormatOps.DisplayList(interpreter.TestConstraints) : FormatOps.DisplayNull);

                    if (empty || ((interpreter.TestFailures != null) && (interpreter.TestFailures.Count > 0)))
                        list.Add("Failures", (interpreter.TestFailures != null) ?
                            interpreter.TestFailures.Count.ToString() : FormatOps.DisplayNull);
                }

                if (list.Count == headerCount)
                    list.Add(FormatOps.DisplayEmpty);
            }
            else
            {
                list.Add(FormatOps.DisplayNull);
            }

            if (OutputStyle == OutputStyle.Boxed)
                return WriteBox(TestInfoBoxName, list, null, false, true,
                    ref hostLeft, ref hostTop, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.Formatted)
                return Write(list, newLine, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.None)
                return true;
            else
                return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteTokenInfo(
            Interpreter interpreter,
            IToken token,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteTokenInfo(interpreter, token, empty, newLine,
                TokenInfoForegroundColor, TokenInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteTokenInfo(
            Interpreter interpreter,
            IToken token,
            bool empty, /* NOT USED */
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            StringPairList list = new StringPairList();

            list.Add("Token");
            list.Add((IPair<string>)null);

            if (token != null)
                list.Add(token.ToList());
            else
                list.Add(FormatOps.DisplayNull);

            if (OutputStyle == OutputStyle.Boxed)
                return WriteBox(TraceInfoBoxName, list, null, false, true,
                    ref hostLeft, ref hostTop, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.Formatted)
                return Write(list, newLine, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.None)
                return true;
            else
                return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteTraceInfo(
            Interpreter interpreter,
            ITraceInfo traceInfo,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteTraceInfo(interpreter, traceInfo, empty, newLine,
                TraceInfoForegroundColor, TraceInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteTraceInfo(
            Interpreter interpreter,
            ITraceInfo traceInfo,
            bool empty, /* NOT USED */
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            StringPairList list = new StringPairList();

            if (traceInfo != null)
            {
                list.Add(traceInfo.ToStringPairList());
            }
            else
            {
                list.Add("TraceInfo");
                list.Add((IPair<string>)null);
                list.Add(FormatOps.DisplayNull);
            }

            if (OutputStyle == OutputStyle.Boxed)
                return WriteBox(TraceInfoBoxName, list, null, false, true,
                    ref hostLeft, ref hostTop, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.Formatted)
                return Write(list, newLine, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.None)
                return true;
            else
                return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteVariableInfo(
            Interpreter interpreter,
            IVariable variable,
            bool elements,
            bool links,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteVariableInfo(interpreter, variable, elements, links, empty, newLine,
                VariableInfoForegroundColor, VariableInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteVariableInfo(
            Interpreter interpreter,
            IVariable variable,
            bool elements,
            bool links,
            bool empty, /* NOT USED */
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            StringPairList list = new StringPairList();

            list.Add("Variable");
            list.Add((IPair<string>)null);

            if (variable != null)
            {
                if (!BuildVariableInfoList(interpreter, variable, elements, links, empty, ref list))
                    list.Add(FormatOps.DisplayUnknown);
            }
            else
            {
                list.Add(FormatOps.DisplayNull);
            }

            if (OutputStyle == OutputStyle.Boxed)
                return WriteBox(VariableInfoBoxName, list, null, false, true,
                    ref hostLeft, ref hostTop, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.Formatted)
                return Write(list, newLine, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.None)
                return true;
            else
                return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteObjectInfo(
            Interpreter interpreter,
            IObject @object,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteObjectInfo(interpreter, @object, empty, newLine,
                ObjectInfoForegroundColor, ObjectInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteObjectInfo(
            Interpreter interpreter,
            IObject @object,
            bool empty,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            StringPairList list = new StringPairList();

            list.Add("Object");
            list.Add((IPair<string>)null);

            if (@object != null)
            {
                if (!BuildObjectInfoList(interpreter, @object, empty, ref list))
                    list.Add(FormatOps.DisplayUnknown);
            }
            else
            {
                list.Add(FormatOps.DisplayNull);
            }

            if (OutputStyle == OutputStyle.Boxed)
                return WriteBox(ObjectInfoBoxName, list, null, false, true,
                    ref hostLeft, ref hostTop, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.Formatted)
                return Write(list, newLine, foregroundColor, backgroundColor);
            else if (OutputStyle == OutputStyle.None)
                return true;
            else
                return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteComplaintInfo(
            Interpreter interpreter,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteComplaintInfo(interpreter, empty, newLine,
                ComplaintInfoForegroundColor, ComplaintInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteComplaintInfo(
            Interpreter interpreter,
            bool empty,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            int count = 0;

            try
            {
                string complaint = null;
                StringPairList list = null;

                //
                // BUGFIX: We never use the colors specified by the caller
                //         (e.g. "Red" / "Black") unless there is actually
                //         a bona fide complaint; therefore, start out with
                //         the system default colors (e.g. "None" / "None").
                //
                ConsoleColor localForegroundColor = DefaultForegroundColor;
                ConsoleColor localBackgroundColor = DefaultBackgroundColor;

                //
                // NOTE: Grab the complaint from the interpreter and check
                //       to see if it is valid (non-null/empty).  We will
                //       display it even if it is empty if we have been
                //       instructed to do so by the caller.  These tests
                //       must be performed in exactly this order because
                //       the first one has the potential for side-effects.
                //
                if (HasComplaint(interpreter, ref complaint) || empty)
                {
                    count++; /* NOTE: Yes, the complaint is valid. */

                    //
                    // NOTE: Is there a valid complaint?  If so, use the
                    //       colors specified by the caller; otherwise,
                    //       use the colors already setup (see above).
                    //
                    if (!String.IsNullOrEmpty(complaint))
                    {
                        localForegroundColor = foregroundColor;
                        localBackgroundColor = backgroundColor;
                    }

                    list = new StringPairList("Complaint", null,
                        FormatOps.DisplayString(complaint));
                }

                if (list != null)
                {
                    if (OutputStyle == OutputStyle.Boxed)
                        return WriteBox(ComplaintInfoBoxName, list, null, false, true,
                            ref hostLeft, ref hostTop, localForegroundColor, localBackgroundColor) &&
                            (++count > 0); /* NOTE: Yes, the complaint was displayed. */
                    else if (OutputStyle == OutputStyle.Formatted)
                        return Write(list, newLine, localForegroundColor, localBackgroundColor) &&
                            (++count > 0); /* NOTE: Yes, the complaint was displayed. */
                    else if (OutputStyle == OutputStyle.None)
                        return true;
                    else
                        return false;
                }
                else
                {
                    return true;
                }
            }
            finally
            {
                //
                // NOTE: *FAIL-SAFE* Only clear the previously stored complaint
                //       if it was valid AND we actually managed to display it.
                //
                if ((count >= 2) && (interpreter != null))
                    interpreter.Complaint = null;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if HISTORY
        public virtual bool WriteHistoryInfo(
            Interpreter interpreter,
            IHistoryFilter historyFilter,
            bool empty,  /* NOT USED */
            bool newLine
            )
        {
            CheckDisposed();

            return WriteHistoryInfo(
                interpreter, historyFilter, empty, newLine, HistoryInfoForegroundColor,
                HistoryInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteHistoryInfo(
            Interpreter interpreter,
            IHistoryFilter historyFilter,
            bool empty,   /* NOT USED */
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            if (interpreter != null)
            {
                StringPairList list = new StringPairList();

                list.Add("History");
                list.Add((IPair<string>)null);

                int headerCount = list.Count;

                lock (interpreter.SyncRoot)
                {
                    if (interpreter.HasHistory())
                    {
                        ClientDataList history = interpreter._History;

                        if (history != null)
                        {
                            foreach (IClientData clientData in history)
                            {
                                if ((clientData != null) &&
                                    RuntimeOps.MatchHistoryData(clientData, historyFilter))
                                {
                                    StringPair pair = FormatOps.HistoryItem(
                                        list.Count - headerCount + 1, clientData, false,
                                        false);

                                    if (pair != null)
                                    {
                                        list.Add(pair);

                                        if ((list.Count - headerCount) == HistoryLimit)
                                            break;
                                    }
                                }
                            }
                        }
                    }
                }

                if (list.Count == headerCount)
                    list.Add(FormatOps.DisplayEmpty);

                if (OutputStyle == OutputStyle.Boxed)
                    return WriteBox(HistoryInfoBoxName, list, null, false, true,
                        ref hostLeft, ref hostTop, foregroundColor, backgroundColor);
                else if (OutputStyle == OutputStyle.Formatted)
                    return Write(list, newLine, foregroundColor, backgroundColor);
                else if (OutputStyle == OutputStyle.None)
                    return true;
                else
                    return false;
            }
            else
            {
                return true;
            }
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteCustomInfo(
            Interpreter interpreter,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteCustomInfo(interpreter, empty, newLine,
                CustomInfoForegroundColor, CustomInfoBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool WriteCustomInfo(
            Interpreter interpreter,
            bool empty,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteAllResultInfo(
            ReturnCode code,
            Result result,
            int errorLine,
            Result previousResult,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteAllResultInfo(code, result, errorLine, previousResult,
                empty, newLine, ResultForegroundColor, ResultBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteAllResultInfo(
            ReturnCode code,
            Result result,
            int errorLine,
            Result previousResult,
            bool empty,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            int count = 0;

            if (empty || (code != ReturnCode.Ok) || (result != null))
            {
                ConsoleColor localForegroundColor = foregroundColor;
                ConsoleColor localBackgroundColor = backgroundColor;

                string formatted;

                if (result != null)
                {
                    formatted = FormatResult(
                        null, code, result, errorLine, Exceptions,
                        Ellipsis, ReplaceNewLines, true,
                        ref localForegroundColor, ref localBackgroundColor);
                }
                else
                {
                    GetResultColors(
                        code, result, ref localForegroundColor,
                        ref localBackgroundColor);

                    formatted = FormatOps.DisplayNull;
                }

                StringPairList list = new StringPairList();

                list.Add("Result", formatted);
                list.Add("Flags", (result != null) ?
                    result.Flags.ToString() : ResultFlags.None.ToString());

                if (WriteLine(list.ToString(), localForegroundColor, localBackgroundColor))
                    count++;
                else
                    return false;

                list.Clear();
            }

            if (empty || (previousResult != null))
            {
                ConsoleColor localForegroundColor = foregroundColor;
                ConsoleColor localBackgroundColor = backgroundColor;

                string formatted;

                if (previousResult != null)
                    formatted = FormatResult(
                        null, previousResult.ReturnCode, previousResult,
                        previousResult.ErrorLine, Exceptions, Ellipsis,
                        ReplaceNewLines, true, ref localForegroundColor,
                        ref localBackgroundColor);
                else
                    formatted = FormatOps.DisplayNull;

                StringPairList list = new StringPairList();

                list.Add("PreviousResult", formatted);
                list.Add("Flags", (previousResult != null) ?
                    previousResult.Flags.ToString() : ResultFlags.None.ToString());

                if (WriteLine(list.ToString(), localForegroundColor, localBackgroundColor))
                    count++;
                else
                    return false;

                list.Clear();
            }

            if (newLine && (count > 0) && !WriteLine())
                return false;

            return true;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteResultInfo(
            string name,
            ReturnCode code,
            Result result,
            int errorLine,
            bool empty,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteResultInfo(name, code, result, errorLine, empty, newLine,
                ResultForegroundColor, ResultBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteResultInfo(
            string name,
            ReturnCode code,
            Result result,
            int errorLine,
            bool empty,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            StringPairList list = new StringPairList();

            if (name == null)
                name = ResultInfoBoxName;

            list.Add(name);
            list.Add((IPair<string>)null);

            int headerCount = list.Count;

            if (empty || (result != null))
                list.Add("Result", FormatOps.DisplayResult(result, false, false));

            if (empty || (code != ReturnCode.Ok))
                list.Add("ReturnCode", code.ToString());

            if (empty || (errorLine != 0))
                list.Add("ErrorLine", errorLine.ToString());

            ResultFlags flags = (result != null) ? result.Flags : ResultFlags.None;

            if (empty || (flags != ResultFlags.None))
                list.Add("Flags", flags.ToString());

            ConsoleColor localForegroundColor = foregroundColor;
            ConsoleColor localBackgroundColor = backgroundColor;

            GetResultColors(code, result, ref localForegroundColor,
                ref localBackgroundColor);

            if (list.Count == headerCount)
                list.Add(FormatOps.DisplayEmpty);

            if (OutputStyle == OutputStyle.Boxed)
                return WriteBox(name, list, null, false, true,
                    ref hostLeft, ref hostTop, localForegroundColor, localBackgroundColor);
            else if (OutputStyle == OutputStyle.Formatted)
                return Write(list, newLine, localForegroundColor, localBackgroundColor);
            else if (OutputStyle == OutputStyle.None)
                return true;
            else
                return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual void WriteHeader(
            Interpreter interpreter,
            ReturnCode code,
            BreakpointType breakpointType,
            string breakpointName,
            IToken token,
            ITraceInfo traceInfo,
            EngineFlags engineFlags,
            SubstitutionFlags substitutionFlags,
            EventFlags eventFlags,
            ExpressionFlags expressionFlags,
            HeaderFlags headerFlags,
            IClientData clientData,
            ArgumentList arguments,
            Result result
            )
        {
            CheckDisposed();

            if (!BeginSection(HeaderSectionName, clientData))
                return;

            bool autoSize = FlagOps.HasFlags(
                headerFlags, HeaderFlags.AutoSize, true);

            bool autoRetry = FlagOps.HasFlags(
                headerFlags, HeaderFlags.AutoRetry, true);

            bool emptySection = FlagOps.HasFlags(
                headerFlags, HeaderFlags.EmptySection, true);

            bool emptyContent = FlagOps.HasFlags(
                headerFlags, HeaderFlags.EmptyContent, true);

            HostFlags hostFlags = autoSize ? GetFlags() : HostFlags.None;

            bool customInfo = FlagOps.HasFlags(
                hostFlags, HostFlags.CustomInfo, true);

            if ((!autoSize || DoesHeaderFit(HeaderFlags.StopPrompt, hostFlags)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.StopPrompt, true))
            {
                WriteLine(StopPrompt, HeaderForegroundColor, HeaderBackgroundColor);
            }

            if ((!autoSize || DoesHeaderFit(HeaderFlags.AnnouncementInfo, hostFlags)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.AnnouncementInfo, true))
            {
                WriteAnnouncementInfo(interpreter,
#if DEBUGGER
                    FlagOps.HasFlags(headerFlags, HeaderFlags.Debug, true) ?
                        GlobalState.GetDebuggerName() :
                        GlobalState.GetPackageName(false),
#else
                    GlobalState.GetPackageName(false),
#endif
                    true);
            }

            bool positioning = DoesSupportPositioning();

            if (positioning)
                GetPosition(ref hostLeft, ref hostTop);

            //
            // NOTE: Make sure we are positioned at the far left.
            //
            if (hostLeft > 0)
            {
                hostLeft = 0;
                hostTop++;

                if (positioning)
                    SetPosition(hostLeft, hostTop);
                else
                    WriteLine();
            }

            int minimumLeft = hostLeft;
            int maximumLeft = minimumLeft;
            int minimumTop = hostTop;
            int maximumTop = minimumTop;
            int savedTop = _Position.Invalid;
            int count = 0;

#if DEBUGGER
            if ((!autoSize || DoesHeaderFit(HeaderFlags.DebuggerInfo, hostFlags)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.DebuggerInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteDebuggerInfo(interpreter, false, emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteDebuggerInfo(interpreter, false, emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }
#endif

            if ((!autoSize || DoesHeaderFit(HeaderFlags.EngineInfo, hostFlags)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.EngineInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteEngineInfo(interpreter, emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteEngineInfo(interpreter, emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }

            if ((!autoSize || DoesHeaderFit(HeaderFlags.ControlInfo, hostFlags)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.ControlInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteControlInfo(interpreter, emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteControlInfo(interpreter, emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }

            if ((!autoSize || DoesHeaderFit(HeaderFlags.EntityInfo, hostFlags)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.EntityInfo, true) &&
                RuntimeOps.IsSameAppDomain(interpreter)) // NOTE: Non-public members (mostly).
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteEntityInfo(interpreter, emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteEntityInfo(interpreter, emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }

            if ((!autoSize || DoesHeaderFit(HeaderFlags.StackInfo, hostFlags)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.StackInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteStackInfo(interpreter, emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteStackInfo(interpreter, emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }

            if ((!autoSize || DoesHeaderFit(HeaderFlags.FlagInfo, hostFlags)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.FlagInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteFlagInfo(interpreter, engineFlags,
                        substitutionFlags, eventFlags, expressionFlags,
                        headerFlags, emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteFlagInfo(interpreter, engineFlags,
                        substitutionFlags, eventFlags, expressionFlags,
                        headerFlags, emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }

            if ((!autoSize || DoesHeaderFit(HeaderFlags.HostInfo, hostFlags)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.HostInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteHostInfo(interpreter, false, false, false, false,
                        emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteHostInfo(interpreter, false, false, false, false,
                        emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }

            if ((!autoSize || DoesHeaderFit(HeaderFlags.ArgumentInfo, hostFlags)) &&
                (emptySection || (breakpointType != BreakpointType.None) || (arguments != null)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.ArgumentInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteArgumentInfo(interpreter, code, breakpointType,
                        breakpointName, arguments, result, emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteArgumentInfo(interpreter, code, breakpointType,
                        breakpointName, arguments, result, emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }

            if ((!autoSize || DoesHeaderFit(HeaderFlags.TestInfo, hostFlags)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.TestInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteTestInfo(interpreter, emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteTestInfo(interpreter, emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }

            if ((!autoSize || DoesHeaderFit(HeaderFlags.TokenInfo, hostFlags)) &&
                (emptySection || (token != null)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.TokenInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteTokenInfo(interpreter, token, emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteTokenInfo(interpreter, token, emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }

            if ((!autoSize || DoesHeaderFit(HeaderFlags.TraceInfo, hostFlags)) &&
                (emptySection || (traceInfo != null)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.TraceInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteTraceInfo(interpreter, traceInfo, emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteTraceInfo(interpreter, traceInfo, emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }

            IVariable variable = (traceInfo != null) ? traceInfo.Variable : null;

            if ((!autoSize || DoesHeaderFit(HeaderFlags.VariableInfo, hostFlags)) &&
                (emptySection || (variable != null)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.VariableInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteVariableInfo(interpreter, variable, false, true,
                        emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteVariableInfo(interpreter, variable, false, true,
                        emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }

            IObject @object = GetObjectFromValue(interpreter, variable);

            if ((!autoSize || DoesHeaderFit(HeaderFlags.ObjectInfo, hostFlags)) &&
                (emptySection || (@object != null)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.ObjectInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteObjectInfo(interpreter, @object, emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteObjectInfo(interpreter, @object, emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }

            if ((!autoSize || DoesHeaderFit(HeaderFlags.InterpreterInfo, hostFlags)) &&
                (emptySection || (interpreter != null)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.InterpreterInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteInterpreterInfo(interpreter, emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteInterpreterInfo(interpreter, emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }

            if ((!autoSize || DoesHeaderFit(HeaderFlags.ComplaintInfo, hostFlags)) &&
                (emptySection || HasComplaint(interpreter)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.ComplaintInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteComplaintInfo(interpreter,
                        emptySection || emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteComplaintInfo(interpreter,
                        emptySection || emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }

            if ((!autoSize || DoesHeaderFit(HeaderFlags.CustomInfo, hostFlags)) &&
                customInfo && /* NOTE: Does the host support CustomInfo? */
                FlagOps.HasFlags(headerFlags, HeaderFlags.CustomInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteCustomInfo(interpreter, emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteCustomInfo(interpreter, emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }

            if ((!autoSize || DoesHeaderFit(HeaderFlags.CallStackInfo, hostFlags)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.CallStackInfo, true) &&
                RuntimeOps.IsSameAppDomain(interpreter)) // NOTE: Non-serializable property.
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteCallStackInfo(interpreter,
                        (interpreter != null) ? interpreter.CallStack : null,
                        FlagOps.HasFlags(headerFlags, HeaderFlags.AllFrames, true),
                        emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteCallStackInfo(interpreter,
                        (interpreter != null) ? interpreter.CallStack : null,
                        FlagOps.HasFlags(headerFlags, HeaderFlags.AllFrames, true),
                        emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }

            if ((!autoSize || DoesHeaderFit(HeaderFlags.ResultInfo, hostFlags)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.ResultInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                if (!WriteResultInfo(null, code, result,
                        (interpreter != null) ? interpreter.ErrorLine : 0,
                        emptyContent, true) && autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteResultInfo(null, code, result,
                        (interpreter != null) ? interpreter.ErrorLine : 0,
                        emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }

#if PREVIOUS_RESULT
            if ((!autoSize || DoesHeaderFit(HeaderFlags.PreviousResultInfo, hostFlags)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.PreviousResultInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                Result previousResult = (interpreter != null) ?
                    interpreter.PreviousResult : null;

                if (!WriteResultInfo(PreviousResultInfoBoxName, (previousResult != null) ?
                        previousResult.ReturnCode : ReturnCode.Ok, previousResult,
                        (previousResult != null) ? previousResult.ErrorLine : 0,
                        emptyContent, true) && autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteResultInfo(PreviousResultInfoBoxName, (previousResult != null) ?
                        previousResult.ReturnCode : ReturnCode.Ok, previousResult,
                        (previousResult != null) ? previousResult.ErrorLine : 0,
                        emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }
#endif

#if HISTORY
            if ((!autoSize || DoesHeaderFit(HeaderFlags.HistoryInfo, hostFlags)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.HistoryInfo, true))
            {
                BeginSection((count + 1) == SectionsPerRow, minimumLeft,
                    ref maximumTop, ref savedTop, ref count);

                IHistoryFilter historyFilter = interpreter.HistoryInfoFilter;

                if (historyFilter == null)
                    historyFilter = InteractiveOps.DefaultHistoryInfoFilter;

                if (!WriteHistoryInfo(interpreter,
                        historyFilter, emptyContent, true) &&
                    autoRetry)
                {
                    BeginSection(true, minimumLeft,
                        ref maximumTop, ref savedTop, ref count);

                    WriteHistoryInfo(interpreter,
                        historyFilter, emptyContent, true);
                }

                EndSection(ref maximumLeft, ref maximumTop);
            }
#endif

            if (hostLeft != minimumLeft)
                maximumTop++;

            if (positioning)
                SetPosition(minimumLeft, maximumTop);
            else
                WriteLine();

            if ((!autoSize || DoesHeaderFit(HeaderFlags.CallStack, hostFlags)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.CallStack, true) &&
                RuntimeOps.IsSameAppDomain(interpreter)) // NOTE: Non-serializable property.
            {
                WriteCallStack(interpreter,
                    (interpreter != null) ? interpreter.CallStack : null,
                    FlagOps.HasFlags(headerFlags, HeaderFlags.AllFrames, true),
                    emptyContent, true);
            }

            //
            // NOTE: Notify the host implementation that we are finished writing the header.
            //
            /* IGNORED */
            EndSection(HeaderSectionName, clientData);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual void WriteFooter(
            Interpreter interpreter,
            ReturnCode code,
            BreakpointType breakpointType,
            string breakpointName,
            IToken token,
            ITraceInfo traceInfo,
            EngineFlags engineFlags,
            SubstitutionFlags substitutionFlags,
            EventFlags eventFlags,
            ExpressionFlags expressionFlags,
            HeaderFlags headerFlags,
            IClientData clientData,
            ArgumentList arguments,
            Result result
            )
        {
            CheckDisposed();

            if (!BeginSection(FooterSectionName, clientData))
                return;

            bool autoSize = FlagOps.HasFlags(
                headerFlags, HeaderFlags.AutoSize, true);

            HostFlags hostFlags = autoSize ? GetFlags() : HostFlags.None;

            if ((!autoSize || DoesHeaderFit(HeaderFlags.GoPrompt, hostFlags)) &&
                FlagOps.HasFlags(headerFlags, HeaderFlags.GoPrompt, true))
            {
                WriteLine(GoPrompt, FooterForegroundColor, FooterBackgroundColor);
            }

            //
            // NOTE: Notify the host implementation that we are finished writing the footer.
            //
            /* IGNORED */
            EndSection(FooterSectionName, clientData);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IHost Members
        private string profile = null;
        public virtual string Profile
        {
            get { CheckDisposed(); return profile; }
            set { CheckDisposed(); profile = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool noColor;
        public virtual bool NoColor
        {
            get { CheckDisposed(); return noColor; }
            set { CheckDisposed(); noColor = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool noTitle;
        public virtual bool NoTitle
        {
            get { CheckDisposed(); return noTitle; }
            set { CheckDisposed(); noTitle = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool noIcon;
        public virtual bool NoIcon
        {
            get { CheckDisposed(); return noIcon; }
            set { CheckDisposed(); noIcon = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool noProfile;
        public virtual bool NoProfile
        {
            get { CheckDisposed(); return noProfile; }
            set { CheckDisposed(); noProfile = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool noCancel;
        public virtual bool NoCancel
        {
            get { CheckDisposed(); return noCancel; }
            set { CheckDisposed(); noCancel = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string defaultTitle;
        public virtual string DefaultTitle
        {
            get { CheckDisposed(); return defaultTitle; }
            set { CheckDisposed(); defaultTitle = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool echo;
        public virtual bool Echo
        {
            get { CheckDisposed(); return echo; }
            set { CheckDisposed(); echo = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool Beep(
            int frequency,
            int duration
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool IsIdle(); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool Clear(); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool ResetFlags()
        {
            CheckDisposed();

            hostFlags = HostFlags.Invalid;

            return true;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool ResetPosition()
        {
            CheckDisposed();

            hostLeft = 0;
            hostTop = 0;

            return true;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool GetPosition(
            ref int left,
            ref int top
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool SetPosition(
            int left,
            int top
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool GetDefaultPosition(
            ref int left,
            ref int top
            )
        {
            CheckDisposed();

            try
            {
                left = hostLeft; top = hostTop;

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool SetDefaultPosition(
            int left,
            int top
            )
        {
            CheckDisposed();

            try
            {
                hostLeft = left; hostTop = top;

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool ResetSize(); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool GetSize(
            ref int width,
            ref int height
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool SetSize(
            int width,
            int height
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract ReturnCode GetMode(
            ChannelType channelType,
            ref uint mode,
            ref Result error
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract ReturnCode SetMode(
            ChannelType channelType,
            uint mode,
            ref Result error
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool ResetColors(); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool GetColors(
            ref ConsoleColor foregroundColor,
            ref ConsoleColor backgroundColor
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool SetForegroundColor(
            ConsoleColor foregroundColor
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool SetBackgroundColor(
            ConsoleColor backgroundColor
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool SetColors(
            bool foreground,
            bool background,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            return (!foreground || SetForegroundColor(foregroundColor)) &&
                   (!background || SetBackgroundColor(backgroundColor));
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract ReturnCode GetColors(
            string theme, /* RESERVED */
            string name,
            bool foreground,
            bool background,
            ref ConsoleColor foregroundColor,
            ref ConsoleColor backgroundColor,
            ref Result error
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract ReturnCode SetColors(
            string theme, /* RESERVED */
            string name,
            bool foreground,
            bool background,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor,
            ref Result error
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract ReturnCode Open(
            ref Result error
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract ReturnCode Close(
            ref Result error
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract ReturnCode Discard(
            ref Result error
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual ReturnCode Reset(
            ref Result error
            )
        {
            CheckDisposed();

            hostFlags = HostFlags.Invalid;

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool Read(
            ref int value
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool ReadKey(
            bool intercept,
            ref IClientData clientData
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if CONSOLE
        [Obsolete()]
        public abstract bool ReadKey(
            bool intercept,
            ref ConsoleKeyInfo keyInfo
            ); /* PRIMITIVE */
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool Write(
            char value,
            bool newLine
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool Write(
            char value,
            int count
            )
        {
            CheckDisposed();

            return Write(value, count, false);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool Write(
            char value,
            int count,
            bool newLine
            )
        {
            CheckDisposed();

            return Write(value, count, newLine, DefaultForegroundColor, DefaultBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool Write(
            char value,
            int count,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            return WriteCore(HostWriteType.Normal, value, count, newLine,
                foregroundColor, backgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool Write(
            char value,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            return Write(value, 1, false, foregroundColor, backgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool Write(
            string value,
            ConsoleColor foregroundColor
            )
        {
            CheckDisposed();

            return Write(value, false, foregroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool Write(
            string value,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            return Write(value, false, foregroundColor, backgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool Write(
            string value,
            bool newLine
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool Write(
            string value,
            bool newLine,
            ConsoleColor foregroundColor
            )
        {
            CheckDisposed();

            return Write(value, newLine, foregroundColor, DefaultBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool Write(
            string value,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            return WriteCore(HostWriteType.Normal, value, newLine,
                foregroundColor, backgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool Write(
            StringPairList list,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            return Write((list != null) ? list.ToString() : null, newLine, foregroundColor,
                backgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteLine(
            string value,
            ConsoleColor foregroundColor
            )
        {
            CheckDisposed();

            return Write(value, true, foregroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteLine(
            string value,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            return Write(value, true, foregroundColor, backgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool BeginSection(
            string name,
            IClientData clientData
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool EndSection(
            string name,
            IClientData clientData
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool BeginBox(
            string name,
            StringPairList list,
            IClientData clientData
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public abstract bool EndBox(
            string name,
            StringPairList list,
            IClientData clientData
            ); /* PRIMITIVE */

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteBox(
            string name,
            string value,
            IClientData clientData,
            bool newLine,
            bool restore,
            ref int left, /* in, out */
            ref int top   /* in, out */
            )
        {
            CheckDisposed();

            return WriteBox(name, value, clientData, newLine, restore, ref left, ref top,
                DefaultForegroundColor, DefaultBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteBox(
            string name,
            string value,
            IClientData clientData,
            bool newLine,
            bool restore,
            ref int left, /* in, out */
            ref int top,  /* in, out */
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            return WriteBox(name, value, clientData, newLine, restore, ref left, ref top,
                foregroundColor, backgroundColor, BoxForegroundColor, BoxBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteBox(
            string name,
            string value,
            IClientData clientData,
            bool newLine,
            bool restore,
            ref int left, /* in, out */
            ref int top,  /* in, out */
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor,
            ConsoleColor boxForegroundColor,
            ConsoleColor boxBackgroundColor
            )
        {
            CheckDisposed();

            return WriteBox(name, new StringPairList(value), clientData, newLine, restore,
                ref left, ref top, foregroundColor, backgroundColor, boxForegroundColor,
                boxBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteBox(
            string name,
            StringPairList list,
            IClientData clientData,
            bool newLine,
            bool restore,
            ref int left, /* in, out */
            ref int top   /* in, out */
            )
        {
            CheckDisposed();

            return WriteBox(name, list, clientData, newLine, restore, ref left, ref top,
                DefaultForegroundColor, DefaultBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteBox(
            string name,
            StringPairList list,
            IClientData clientData,
            bool newLine,
            bool restore,
            ref int left, /* in, out */
            ref int top,  /* in, out */
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            return WriteBox(name, list, clientData, newLine, restore, ref left, ref top,
                foregroundColor, backgroundColor, BoxForegroundColor, BoxBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public virtual bool WriteBox(
            string name,
            StringPairList list,
            IClientData clientData,
            bool newLine,
            bool restore,
            ref int left, /* in, out */
            ref int top,  /* in, out */
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor,
            ConsoleColor boxForegroundColor,
            ConsoleColor boxBackgroundColor
            )
        {
            CheckDisposed();

            if (list != null)
            {
                //
                // NOTE: Notify the host implementation that we are about to
                //       write a box.  This lets the host implementation know
                //       that all the following writes (i.e. until the call to
                //       EndBox) logically belong to the same composite item.
                //
                if (!BeginBox(name, list, clientData))
                    return false;

                //
                // NOTE: Does the host support sizing and positioning for the
                //       content area?
                //
                bool sizing = DoesSupportSizing();
                bool positioning = DoesSupportPositioning();

                //
                // NOTE: If necessary, save the current host position for later
                //       restoration.
                //
                int savedLeft = _Position.Invalid;
                int savedTop = _Position.Invalid;

                if (positioning && restore && !GetPosition(ref savedLeft, ref savedTop))
                    return false;

                try
                {
                    //
                    // NOTE: What is the margin between the name and value on a
                    //       particular line in the box (if applicable to that
                    //       line)?
                    //
                    int margin = BoxMargin;

                    if (margin < 0)
                        margin = ContentMargin;

                    //
                    // NOTE: What is the overall content length limit?
                    //
                    int limit = ContentWidth - margin;

                    //
                    // NOTE: The length limit must take into account the horizontal
                    //       offset into the display area.
                    //
                    // TODO: Should this always be done unconditionally (i.e. maybe
                    //       make it optional in the future, via a bool argument)?
                    //
                    limit -= left;

                    //
                    // NOTE: Make sure we have enough horizontal space to output
                    //       something meaningful.
                    //
                    if ((limit > 0) && (limit >= ContentThreshold))
                    {
                        //
                        //  NOTE: If necessary, get the total width and height of
                        //        the content area.
                        //
                        int width = 0;
                        int height = 0;

                        if (!sizing || GetSize(ref width, ref height))
                        {
                            //
                            // NOTE: Calculate how much vertical space we need.
                            //
                            int maximumTop = top + 2 + (newLine ? (list.Count * 2) : list.Count);

                            //
                            // NOTE: Make sure we have enough vertical space to output
                            //       the entire box.
                            //
                            if (!sizing || (maximumTop < height))
                            {
                                //
                                // NOTE: Fetch the currently configured box limit (i.e. the maximum
                                //       width allowed for one box).  If no valid box limit is set
                                //       or it exceeds the overall limit, use the overall limit
                                //       instead.
                                //
                                int boxLimit = BoxWidth;

                                if ((boxLimit < 0) || (boxLimit > limit))
                                    boxLimit = limit;

                                //
                                // NOTE: Remove the content margin from the box limit.  We do this
                                //       because we need to calculate the maximum string length
                                //       (below) without taking into the account the margin.  The
                                //       value written will be truncated to this limit and ellipsed
                                //       as necessary.
                                //
                                boxLimit -= margin;

                                //
                                // NOTE: Make sure we can still output something.
                                //
                                if (boxLimit >= 0)
                                {
                                    //
                                    // NOTE: Figure out the maximum length for any item in the supplied
                                    //       list that will fit within the physical bounds of the content
                                    //       area.
                                    //
                                    int length = ListOps.GetMaximumLength(list, NameValueFormat, boxLimit);

                                    //
                                    // NOTE: Add a margin for ease of reading.  These spaces, if any,
                                    //       will be used for padding in between the name and the value
                                    //       of each name/value pair.
                                    //
                                    length += margin;

                                    //
                                    // NOTE: Use the caller's position variables as the starting point
                                    //       for the box.
                                    //
                                    int newLeft = left;
                                    int newTop = top;

                                    //
                                    // NOTE: Grab the "character set" (i.e. string) that we are going
                                    //       to use to draw the pieces of the box.
                                    //
                                    string characterSet = BoxCharacterSets[BoxCharacterSet];

                                    //
                                    // NOTE: Make sure the configured character set includes all the
                                    //       characters we need.
                                    //
                                    if (characterSet.Length < ((int)BoxCharacter.Space + 1))
                                        return false;

                                    //
                                    // NOTE: Set the current position to the initial position of the
                                    //       top line of the box (i.e. the initial position of the box).
                                    //
                                    if (positioning && !SetPosition(newLeft, newTop))
                                        return false;
                                    else if (!positioning && !WriteLine())
                                        return false;

                                    //
                                    // NOTE: Draw the upper left corner of the box.
                                    //
                                    if (!Write(characterSet[(int)BoxCharacter.TopLeft],
                                            boxForegroundColor, boxBackgroundColor))
                                    {
                                        return false;
                                    }

                                    //
                                    // NOTE: Draw the upper middle section of the box.
                                    //
                                    if (!Write(characterSet[(int)BoxCharacter.Horizontal],
                                            length, false, boxForegroundColor,
                                            boxBackgroundColor))
                                    {
                                        return false;
                                    }

                                    //
                                    // NOTE: Draw the upper right corner of the box.
                                    //
                                    if (!Write(characterSet[(int)BoxCharacter.TopRight],
                                            boxForegroundColor, boxBackgroundColor))
                                    {
                                        return false;
                                    }

                                    //
                                    // NOTE: We just wrote a line, advance the line counter.
                                    //
                                    newTop++;

                                    foreach (IPair<string> element in list)
                                    {
                                        IToString toString = element as IToString;

                                        if ((element != null) && (toString != null))
                                        {
                                            string value;

                                            //
                                            // NOTE: Check to see if this is really a name/value pair or
                                            //       just a single value.
                                            //
                                            if ((element.X != null) && (element.Y != null))
                                            {
                                                //
                                                // NOTE: Start with the default format.
                                                //
                                                string format = NameValueFormat;

                                                //
                                                // NOTE: Grab and format the content of the box for this
                                                //       line.
                                                //
                                                value = StringOps.NormalizeWhiteSpace(
                                                    toString.ToString(format, boxLimit, true),
                                                    Characters.Space, false, false);

                                                //
                                                // NOTE: Recalculate the layout to justify the name to
                                                //       the left and the value to the right.
                                                //
                                                format = format.Replace(
                                                    Characters.Space.ToString(),
                                                    StringOps.StrRepeat(
                                                        (length - value.Length) + 1,
                                                        characterSet[(int)BoxCharacter.Space]));

                                                //
                                                // NOTE: Reformat the content of the box for this line
                                                //       with the name and value left and right justified.
                                                //
                                                value = StringOps.NormalizeWhiteSpace(
                                                    toString.ToString(format, boxLimit + margin, true),
                                                    Characters.Space, false, false);
                                            }
                                            else if (element.X != null)
                                            {
                                                value = StringOps.NormalizeWhiteSpace(
                                                    element.X, Characters.Space, false, false);

                                                //
                                                // NOTE: Truncate the content to fit within the actual
                                                //       physical limits of the content area.
                                                //
                                                if (value.Length > boxLimit)
                                                    value = FormatOps.Ellipsis(value, boxLimit, true);

                                                //
                                                // NOTE: Center pad the content of the box for this line
                                                //       with spaces to the maximum length of any of the
                                                //       content.
                                                //
                                                value = StringOps.PadCenter(
                                                    value, length, characterSet[(int)BoxCharacter.Space]);
                                            }
                                            else if (element.Y != null)
                                            {
                                                value = StringOps.NormalizeWhiteSpace(
                                                    element.Y, Characters.Space, false, false);

                                                //
                                                // NOTE: Truncate the content to fit within the actual
                                                //       physical limits of the content area.
                                                //
                                                if (value.Length > boxLimit)
                                                    value = FormatOps.Ellipsis(value, boxLimit, true);

                                                //
                                                // NOTE: Center pad the content of the box for this line
                                                //       with spaces to the maximum length of any of the
                                                //       content.
                                                //
                                                value = StringOps.PadCenter(
                                                    value, length, characterSet[(int)BoxCharacter.Space]);
                                            }
                                            else
                                            {
                                                value = null;
                                            }

                                            //
                                            // NOTE: Do we need to output anything for this content line
                                            //       of the box?
                                            //
                                            if (value != null)
                                            {
                                                //
                                                // NOTE: Set the current position to the initial position
                                                //       of the current line of the box.
                                                //
                                                if (positioning && !SetPosition(newLeft, newTop))
                                                    return false;
                                                else if (!positioning && !WriteLine())
                                                    return false;

                                                //
                                                // NOTE: Draw the left side of the box for this line.
                                                //
                                                if (!Write(characterSet[(int)BoxCharacter.Vertical],
                                                        boxForegroundColor, boxBackgroundColor))
                                                {
                                                    return false;
                                                }

                                                //
                                                // NOTE: Draw the content of the box for this line.
                                                //
                                                if (!Write(value, foregroundColor, backgroundColor))
                                                    return false;

                                                //
                                                // NOTE: Draw the right side of the box for this line.
                                                //
                                                if (!Write(characterSet[(int)BoxCharacter.Vertical],
                                                        boxForegroundColor, boxBackgroundColor))
                                                {
                                                    return false;
                                                }

                                                //
                                                // NOTE: We just wrote another line, advance the line
                                                //       counter.
                                                //
                                                newTop++;
                                            }
                                        }
                                        else
                                        {
                                            //
                                            // NOTE: Set the current position to the initial position
                                            //       of the current line of the box.
                                            //
                                            if (positioning && !SetPosition(newLeft, newTop))
                                                return false;
                                            else if (!positioning && !WriteLine())
                                                return false;

                                            //
                                            // NOTE: Draw the left side junction of the box for this
                                            //       line.
                                            //
                                            if (!Write(characterSet[(int)BoxCharacter.LeftJunction],
                                                    boxForegroundColor, boxBackgroundColor))
                                            {
                                                return false;
                                            }

                                            //
                                            // NOTE: Draw a horizontal line.
                                            //
                                            if (!Write(characterSet[(int)BoxCharacter.Horizontal],
                                                    length, false, boxForegroundColor,
                                                    boxBackgroundColor))
                                            {
                                                return false;
                                            }

                                            //
                                            // NOTE: Draw the right side junction of the box for this
                                            //       line.
                                            //
                                            if (!Write(characterSet[(int)BoxCharacter.RightJunction],
                                                    boxForegroundColor, boxBackgroundColor))
                                            {
                                                return false;
                                            }

                                            //
                                            // NOTE: We just wrote another line, advance the line
                                            //       counter.
                                            //
                                            newTop++;
                                        }

                                        if (newLine)
                                        {
                                            //
                                            // NOTE: Set the current position to the initial position
                                            //       of the next line of the box (the line counter was
                                            //       just advanced, above).
                                            //
                                            if (positioning && !SetPosition(newLeft, newTop))
                                                return false;
                                            else if (!positioning && !WriteLine())
                                                return false;

                                            //
                                            // NOTE: Draw the left side of the box for this blank line.
                                            //
                                            if (!Write(characterSet[(int)BoxCharacter.Vertical],
                                                    boxForegroundColor, boxBackgroundColor))
                                            {
                                                return false;
                                            }

                                            //
                                            // NOTE: Draw a blank line.
                                            //
                                            if (!Write(characterSet[(int)BoxCharacter.Space],
                                                    length, false, foregroundColor,
                                                    backgroundColor))
                                            {
                                                return false;
                                            }

                                            //
                                            // NOTE: Draw the right side of the box for this blank line.
                                            //
                                            if (!Write(characterSet[(int)BoxCharacter.Vertical],
                                                    boxForegroundColor, boxBackgroundColor))
                                            {
                                                return false;
                                            }

                                            //
                                            // NOTE: We just wrote another line, advance the line
                                            //       counter.
                                            //
                                            newTop++;
                                        }
                                    }

                                    //
                                    // NOTE: Set the current position to the initial position
                                    //       of the bottom line of the box.
                                    //
                                    if (positioning && !SetPosition(newLeft, newTop))
                                        return false;
                                    else if (!positioning && !WriteLine())
                                        return false;

                                    //
                                    // NOTE: Draw the lower left corner of the box.
                                    //
                                    if (!Write(characterSet[(int)BoxCharacter.BottomLeft],
                                            boxForegroundColor, boxBackgroundColor))
                                    {
                                        return false;
                                    }

                                    //
                                    // NOTE: Draw the lower middle section of the box.
                                    //
                                    if (!Write(characterSet[(int)BoxCharacter.Horizontal],
                                            length, false, boxForegroundColor,
                                            boxBackgroundColor))
                                    {
                                        return false;
                                    }

                                    //
                                    // NOTE: Draw the lower right corner of the box.
                                    //
                                    if (!Write(characterSet[(int)BoxCharacter.BottomRight],
                                            boxForegroundColor, boxBackgroundColor))
                                    {
                                        return false;
                                    }

                                    //
                                    // NOTE: We just wrote another line, advance the line
                                    //       counter.
                                    //
                                    newTop++;

                                    //
                                    // NOTE: Update the horizontal position as well (so that
                                    //       the caller knows where we left off).
                                    //
                                    newLeft += (length + 2);

                                    //
                                    // NOTE: Update the caller's variables with the new
                                    //       positions.
                                    //
                                    left = newLeft;
                                    top = newTop;

                                    //
                                    // NOTE: If we get to this point then we have totally
                                    //       succeeded.
                                    //
                                    return true;
                                }
                            }
                        }
                    }

#if false
                    DebugOps.Complain(ReturnCode.Error,
                        "cannot write box, insufficient space");
#endif

                    //
                    // NOTE: We cannot output anything because there is no space
                    //       (either horizontal, vertical, or both) to do so after
                    //       taking into account the amount of content the caller
                    //       requests us to write and the physical bounds of the
                    //       content area.
                    //
                    /* FALL-THROUGH */
                }
                catch
                {
                    // do nothing.
                }
                finally
                {
                    //
                    // NOTE: Restore the previously saved position?
                    //
                    if (positioning && restore)
                        /* IGNORED */
                        SetPosition(savedLeft, savedTop);

                    //
                    // NOTE: Notify the host implementation that we are
                    //       finished writing the box.
                    //
                    /* IGNORED */
                    EndBox(name, list, clientData);
                }
            }

            return false;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IDisposable Members
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IDisposable "Pattern" Members
        private bool disposed;
        private void CheckDisposed() /* throw */
        {
#if THROW_ON_DISPOSED
            if (disposed && Engine.IsThrowOnDisposed(null, false))
                throw new ObjectDisposedException(typeof(Default).Name);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual void Dispose(bool disposing)
        {
            if (!disposed)
            {
                //if (disposing)
                //{
                //    ////////////////////////////////////
                //    // dispose managed resources here...
                //    ////////////////////////////////////
                //}

                //////////////////////////////////////
                // release unmanaged resources here...
                //////////////////////////////////////

                disposed = true;
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Destructor
        ~Default()
        {
            Dispose(false);
        }
        #endregion
    }
}
